"""This module supports loading all the data components required to make
a CRDS lookup table for an instrument.

get_cached_mapping loads the closure of the given context file from the
local CRDS store,  caching the result.

>>> p = get_cached_mapping("hst.pmap")

The initial HST pipeline mappings are self-consistent and there are none
missing:

>>> p.missing_mappings()
[]

There are 72 pmap, imap, and rmap files in the entire HST pipeline:

>>> len(p.mapping_names()) > 50
True

There are over 9000 reference files known to the initial CRDS mappings scraped
from the CDBS HTML table dump:

>>> len(p.reference_names()) > 1000
True

Pipeline reference files are also broken down by instrument:

>>> sorted(p.reference_name_map().keys())
['acs', 'cos', 'nicmos', 'stis', 'wfc3', 'wfpc2']

>>> i = load_mapping("hst_acs.imap")

The ACS instrument has 15 associated mappings,  including the instrument
context:

>>> len(i.mapping_names()) > 10
True

The ACS instrument has 3983 associated reference files in the hst_acs.imap
context:

>>> len(i.reference_names()) > 2000
True

Active instrument references are also broken down by filetype:

>>> sorted(i.reference_name_map()["crrejtab"])
['j4d1435lj_crr.fits', 'lci1518ej_crr.fits', 'lci1518fj_crr.fits', 'n4e12510j_crr.fits', 'n4e12511j_crr.fits']

>>> r = ReferenceMapping.from_file("hst_acs_biasfile.rmap")
>>> len(r.reference_names())  > 500
True
"""
import os.path
import glob
import json

from collections import namedtuple

# ===================================================================

from pkg_resources import Requirement

from . import log, utils, config, selectors, substitutions

# XXX For backward compatability until refactored away.
from .config import locate_file, locate_mapping, locate_reference
from .config import mapping_exists, is_mapping

from . import exceptions as crexc
from .custom_dict import LazyFileDict
from .mapping_verifier import MAPPING_VERIFIER
from .log import srepr
from .constants import ALL_OBSERVATORIES, INSTRUMENT_KEYWORDS

# from crds import data_file   (deferred to actual uses)

# ===================================================================

__all__ = [
    "Mapping",
    "ContextMapping",
    "PipelineContext",
    "InstrumentContext",
    "ReferenceMapping",
    "get_cached_mapping",
    "load_mapping",
    "fetch_mapping",
    "asmapping",
    "locate_mapping",
    "list_mappings",
    "list_references",
    "get_best_references",
    "mapping_type",
]

# ===================================================================

Filetype = namedtuple("Filetype","header_keyword,extension,rmap")
Failure  = namedtuple("Failure","header_keyword,message")
Filemap  = namedtuple("Filemap","date,file,comment")

# =============================================================================

class LowerCaseDict(dict):
    """Used to return Mapping header string values uniformly as lower case.

    >>> d = LowerCaseDict([("this","THAT"), ("another", "(ESCAPED)")])

    Ordinarily,  all string values are mapped to lower case:

    >>> d["this"]
    'that'

    Values bracketed by () are returned unaltered in order to support header Python
    expressions which are typically evaluated in the context of an incoming header
    (FITS) dictionary,  all upper case:

    >>> d["another"]
    '(ESCAPED)'
    """
    def __getitem__(self, key):
        val = super(LowerCaseDict, self).__getitem__(key)
        # Return string values as lower case,  but exclude literal expressions surrounded by ()
        # for case-sensitive HST rmap relevance expressions.
        if isinstance(val, str) and not (val.startswith("(") and val.endswith(")")):
            val = val.lower()
        return val

    def get(self, key, default):
        if key in self:
            return self[key]
        else:
            return default

    def __repr__(self):
        """
        >> LowerCaseDict([("this","THAT"), ("another", "(ESCAPED)")])
        LowerCaseDict({'this': 'that', 'another': '(ESCAPED)'})
        """
        return self.__class__.__name__ + "({})".format(repr({ key: self[key] for key in self }))

# ===================================================================

class Mapping:
    """Mapping is the abstract baseclass for PipelineContext,
    InstrumentContext, and ReferenceMapping.
    """
    required_attrs = ["mapping", "parkey", "name", "derived_from"]

    # Strictly speaking,  belongs to ContextMapping, need some default though
    # If False,  only specifies nested Mappings by filename
    specifies_references = False

    # no precursor file if derived_from contains any of these.
    null_derivation_substrings = ("generated", "cloning", "by hand")

    def __init__(self, filename, header, selector, **keys):
        self.filename = filename
        self.header = LowerCaseDict(header)   # consistent lower case values
        self.selector = selector
        self.keys = keys
        # Keys should already be as good as they get
        self.comment = keys.pop("comment", None)
        self.path = keys.get("path", None)
        for name in self.required_attrs:
            if name not in self.header:
                raise crexc.MissingHeaderKeyError(
                    "Required header key " + repr(name) + " is missing.")
        # Header aren't known good/present until checked above.
        self.mapping = self.header["mapping"]
        self.parkey = self.header["parkey"]
        self.extra_keys = tuple(self.header.get("extra_keys", ()))

    def _check_type(self):
        """Verify that the 'mapping' element of the header matches 'self.mapping_type'."""
        assert self.mapping == self.mapping_type, \
            "Expected header mapping='{}' but got mapping='{}' in '{}'".format(
                self.mapping_type, self.mapping.upper(), self.filename)

    def validate_mapping(self):
        """Validate `self` only implementing any checks to be performed by
        crds.certify.   ContextMappings are mostly validated at load time.
        Stick extra checks for context mappings here.
        """
        log.verbose("Validating", repr(self.basename), "with parameters", repr(self.parkey))
        self.validate()

    def validate(self):
        """Perform generic recursive Mapping checks, including definition of all
        required header fields as defined by Mapping class/sub-class.  Since
        this requires loading all sub-mappings, it is factored out to facilitate
        demand based loading elsewhere.

        Other methods of validation are performed when a Mapping is loaded,  including
        sha1sum and acceptability of byte codes,  neither of which requires loading
        sub-mappings.
        """
        self._check_type()

    def check_observatory(self):
        """Verify self.observatory is a supported observatory."""
        assert self.observatory in ALL_OBSERVATORIES, \
            "Invalid observatory " + repr(self.observatory) + " in " + repr(self.filename)

    def check_instrument(self):
        """Verify self.instrument is a supported instrument."""
        assert self.instrument in self.obs_package.INSTRUMENTS, \
            "Invalid instrument " + repr(self.instrument) + " in " + repr(self.filename)

    def check_filekind(self):
        """Verify self.filekind is a supported filekind."""
        assert self.filekind in self.obs_package.FILEKINDS, \
            "Invalid filekind " + repr(self.filekind) + " in " + repr(self.filename)

    def _trace_compare(self, other, show_equal=False):
        utils.trace_compare(self, other, show_equal)
        for key in self.selections:
            self.selections[key]._trace_compare(other.selections[key], show_equal)

    @property
    def basename(self):
        return os.path.basename(self.filename)

    def __repr__(self):
        """A subclass-safe repr which includes required parameters except for
        'mapping' which is implied by the classname.
        """
        rep = self.__class__.__name__ + "("
        rep += repr(self.filename)
        rep += ", "
        rep = rep[:-2] + ")"
        return rep

    def __str__(self):
        """Return the source text of the Mapping."""
        return self.format()

    def __getattr__(self, attr):
        """Enable access to required header parameters as 'self.<parameter>'"""
        if "header" in self.__dict__ and attr in self.header:
            return self.header[attr]   # Note:  header is a class which mutates values,  see LowerCaseDict.
        else:
            raise AttributeError("Invalid or missing header key " + repr(attr))

    @classmethod
    def from_s3(cls, basename, *args, **keys):
        log.verbose("Loading mapping from S3", repr(basename), verbosity=55)
        s3_uri = config.get_uri(basename)
        text = utils.get_uri_content(s3_uri)
        return cls.from_string(text, basename, *args, **keys)

    @classmethod
    def from_file(cls, basename, *args, **keys):
        """Load a mapping file `basename` and do syntax and basic validation.  If `path` is
        specified, recursively load all files relative to `path` and include path in the
        name of the mapping.
        """
        log.verbose("Loading mapping", repr(basename), verbosity=55)
        path = keys.get("path", None)
        if path:
            filename = os.path.join(path, os.path.basename(basename))
            basename = filename
        else:
            filename = config.locate_mapping(basename)
        text = utils.get_uri_content(filename)
        return cls.from_string(text, basename, *args, **keys)

    @classmethod
    def from_string(cls, text, basename="(noname)", *args, **keys):
        """Construct a mapping from string `text` nominally named `basename`."""
        keys.pop("comment", None) #  discard comment if defined
        header, selector, comment = cls._parse_header_selector(text, basename)
        mapping = cls(basename, header, selector, comment=comment, **keys)
        try:
            mapping._check_hash(text)
        except crexc.ChecksumError as exc:
            ignore = keys.get("ignore_checksum", False) or config.get_ignore_checksum()
            if ignore == "warn":
                log.warning("Checksum error", ":", str(exc))
            elif ignore:
                pass
            else:
                raise
        return mapping

    @classmethod
    def _parse_header_selector(cls, text, where=""):
        """Given a mapping at `filepath`,  validate it and return a fully
        instantiated (header, selector) tuple.
        """
        with log.augment_exception("Can't load file " + where,
                                   exception_class=crexc.MappingError):
            code = MAPPING_VERIFIER.compile_and_check(text)
            header, selector, comment = cls._interpret(code)
        return LowerCaseDict(header), selector, comment

    @classmethod
    def _interpret(cls, code):
        """Interpret a valid rmap code object and return it's header and
        selector.
        """
        namespace = {}
        namespace.update(selectors.SELECTORS)
        exec(code, namespace)
        header = LowerCaseDict(namespace["header"])
        selector = namespace["selector"]
        comment = namespace.get("comment", None)
        if isinstance(selector, selectors.Parameters):
            return header, selector.instantiate(header), comment
        elif isinstance(selector, dict):
            return header, selector, comment
        else:
            raise crexc.MappingFormatError("selector must be a dict or a Selector.")

    def missing_references(self):
        """Get the references mentioned by the closure of this mapping but not in the cache."""
        return [ ref for ref in self.reference_names() if not config.file_in_cache(self.name, self.observatory) ]

    def missing_mappings(self):
        """Get the mappings mentioned by the closure of this mapping but not in the cache."""
        return [ mapping for mapping in self.mapping_names() if not config.file_in_cache(self.name, self.observatory) ]

    @property
    def locate(self):
        """Return the "locate" module associated with self.observatory."""
        return utils.get_object("crds", self.observatory, "locate")

    @property
    def obs_package(self):
        """Return the package (__init__) associated with self.observatory."""
        return utils.get_observatory_package(self.observatory)

    @property
    def instr_package(self):
        """Return the module associated with self.instrument."""
        return utils.get_object("crds", self.observatory, self.instrument)

    def format(self):
        """Return the string representation of this mapping, i.e. pretty
        serialization.   This is currently limited to initially creating rmaps,
        not rewriting them since it is based on internal representations and
        therefore loses comments.
        """
        if self.comment:
            return "header = {0}\n\ncomment = {1}\n\nselector = {2}\n" .format(
                self._format_header(), self._format_comment(), self._format_selector())
        else:
            return "header = {0}\n\nselector = {1}\n".format(self._format_header(), self._format_selector())

    def _format_dict(self, dict_, indent=0):
        """Return indented source code for nested `dict`."""
        prefix = indent*" "*4
        output = "{\n"
        for key, val in sorted(dict_.items()):
            if isinstance(val, dict):
                rval = self._format_dict(val, indent+1)
            else:
                rval = repr(val)
            output += prefix + " "*4 + repr(key) + " : " + rval + ",\n"
        output += prefix + "}"
        return output

    def _format_header(self):
        """Return the code string for the mapping header."""
        return self._format_dict(self.header)

    def _format_selector(self):
        """Return the code string for the mapping body/selector."""
        if isinstance(self.selector, dict):
            return self._format_dict(self.selector)
        else:
            return self.selector.format()

    def _format_comment(self):
        """Return the string representation of the multi-line comment block."""
        return '"""' + self.comment + '"""'

    def write(self, filename=None):
        """Write out this mapping to the specified `filename`,
        or else self.filename. DOES NOT PRESERVE COMMENTS.
        """
        if filename is None:
            filename = self.filename
        else:
            self.filename = filename
        self.header["sha1sum"] = self._get_checksum(self.format())
        with open(filename, "w+") as handle:
            handle.write(self.format())

    def _check_hash(self, text):
        """Verify that the mapping header has a checksum and that it is
        correct,  else raise an appropriate exception.
        """
        old = self.header.get("sha1sum", None)
        if old is None:
            raise crexc.ChecksumError("sha1sum is missing in " + repr(self.basename))
        if self._get_checksum(text) != self.header["sha1sum"]:
            raise crexc.ChecksumError("sha1sum mismatch in " + repr(self.basename))

    def _get_checksum(self, text):
        """Compute the rmap checksum over the original file contents.  Skip over the sha1sum line."""
        # Compute the new checksum over everything but the sha1sum line.
        # This will fail if sha1sum appears for some other reason.  It won't ;-)
        text = "".join([line for line in text.splitlines(True) if "sha1sum" not in line])
        return utils.str_checksum(text)

    rewrite_checksum = write
    #    """Re-write checksum updates the checksum for a Mapping writing the
    #    result out to `filename`.
    #    """

    @utils.cached
    def get_required_parkeys(self):
        """Determine the set of parkeys required for this mapping and all the mappings selected by it."""
        parkeys = set(self.parkey)
        parkeys |= set(self.header.get("extra_keys",[]))
        if hasattr(self, "selections"):
            for selection in self.selections.normal_values():
                parkeys |= set(selection.get_required_parkeys())
        return sorted(parkeys)

    def minimize_header(self, header):
        """Return only those items of `header` which are required to determine
        bestrefs.   Missing keys are set to 'UNDEFINED'.
        """
        header = self.locate.fits_to_parkeys(header)   # reference vocab --> dataset vocab
        if isinstance(self, PipelineContext):
            instrument = self.get_instrument(header)
            mapping = self.get_imap(instrument)
            keys = mapping.get_required_parkeys() + [self.instrument_key]
        else:
            keys = self.get_required_parkeys()
        minimized = {}
        for key in keys:
            minimized[key] = header.get(key.lower(),
                                        header.get(key.upper(), "UNDEFINED"))
        return minimized

    def get_minimum_header(self, dataset, original_name=None):
        """Return the names and values of `dataset`s header parameters which
        are required to compute best references for it.   `original_name` is
        used to determine file type when `dataset` is a temporary file with a
        useless name.
        """
        from crds import data_file
        header = data_file.get_conditioned_header(dataset, original_name=original_name)
        return self.minimize_header(header)

    def difference(self, new_mapping, path=(), pars=(), include_header_diffs=False, recurse_added_deleted=False):
        """Compare `self` with `new_mapping` and return a list of difference
        tuples,  prefixing each tuple with context `path`.
        """
        new_mapping = asmapping(new_mapping, cache="readonly")
        log.verbose("Difference:", self.filename, "vs.",new_mapping.filename)
        differences = self.difference_header(new_mapping, path=path, pars=pars) if include_header_diffs else []
        for key in self.selections:  # Check for deleted or replaced keys in self / old mapping.
            if key not in new_mapping.selections:   # deletions from self
                diff = selectors.DiffTuple(
                    * path + ((self.filename, new_mapping.filename), (key,),
                    "deleted " + repr(self._value_name(key))),
                    parameter_names = pars + (self.diff_name, self.parkey, "DIFFERENCE",))
                if recurse_added_deleted and self._is_normal_value(key):
                    # Get tuples for all implicitly deleted nested files.
                    nested_diffs = self.selections[key].diff_files("deleted",
                        path = path + ((self.filename,),), pars = pars + (self.diff_name,),)
                else: # either no recursion or key is special and cannot be recursed.
                    nested_diffs = []
            elif self._value_name(key) != new_mapping._value_name(key):
                # replacements in self
                # different basenames identify context-to-context updates
                # different filenames identify cache-to-cache differences
                if self._value_xsum(key) != new_mapping._value_xsum(key):
                    # omit identical files from report of diffs (cache-to-cache)
                    # if contents are different,  include a diff tuple
                    # otherwise assume higher level file at different path
                    diff = selectors.DiffTuple(
                        * (path + ((self.filename, new_mapping.filename), (key,),
                                   "replaced " + repr(self._value_name(key)) + " with " + repr(new_mapping._value_name(key)))),
                        parameter_names = pars + (self.diff_name, self.parkey, "DIFFERENCE",))
                else:
                    # higher levels file identical, nested cache-to-cache files may differ
                    diff = None
                if self._is_normal_value(key) and new_mapping._is_normal_value(key):   # mapping replacements
                    # recursion needed if both selections are mappings.
                    nested_diffs = self.selections[key].difference( new_mapping.selections[key],
                        path = path + ((self.filename, new_mapping.filename,), ), pars = pars + (self.diff_name,),
                        include_header_diffs=include_header_diffs, recurse_added_deleted=recurse_added_deleted)
                elif recurse_added_deleted:  # include added/deleted cases from normal mapping replacing special, vice versa
                    if self._is_normal_value(key):  # new_mapping is special
                        nested_diffs = self.selections[key].diff_files("deleted",
                            path = path + ((self.filename,),), pars = pars + (self.diff_name,),)
                    elif new_mapping._is_normal_value(key):   # self is special
                        nested_diffs = new_mapping.selections[key].diff_files("added",
                            path = path + ((self.filename,),), pars = pars + (self.diff_name,),)
                    else:  # recurse but both special,  handled by basic diff above.
                        nested_diffs = []
                else:  # No recursion,  handled by basic diff above.
                    nested_diffs = []
            else:  # values are the same,  no diff or nested diffs.
                diff = None
                nested_diffs = []
            # Nullify intermediate diffs based solely on path names for cache-to-cache comparisons
            # Cache-to-cache difference is tricky,  because recursion is required but higher level
            # diffs by path-only are uninteresting.
            if self.basename == new_mapping.basename and not nested_diffs:
                diff = None
            if diff:
                differences.append(diff)
            differences.extend(nested_diffs)
        for key in new_mapping.selections:
            if key not in self.selections:      # Additions to self
                diff = selectors.DiffTuple(
                    * path + ((self.filename, new_mapping.filename), (key,),
                    "added " + repr(new_mapping._value_name(key))),
                    parameter_names = pars + (self.diff_name, self.parkey, "DIFFERENCE",))
                differences.append(diff)
                if recurse_added_deleted and new_mapping._is_normal_value(key):
                    # Get tuples for all implicitly added nested files.
                    nested_adds = new_mapping.selections[key].diff_files("added",
                        path = path + ((self.filename,),), pars = pars + (self.diff_name,),)
                    differences.extend(nested_adds)
            else: # replacement case already handled in first for-loop,  not needed in reverse.
                pass
        return sorted(differences)

    def diff_files(self, added_deleted, path=(), pars=()):
        """Return the list of diff tuples for all nested changed files in a higher level addition
        or deletion.   added_deleted should be "added" or "deleted"
        """
        diffs = []
        for key, selection in self.selections.items():
            if self._is_normal_value(key):
                diffs.extend(selection.diff_files(added_deleted, path + (key,), pars + (self.diff_name,)))
            else:
                delete_special = selectors.DiffTuple(
                        * (path + ((self.filename,), (key,),
                        added_deleted + " " + repr(self._value_name(key)))),
                        parameter_names = pars + (self.diff_name, self.parkey, "DIFFERENCE",))
                diffs.append(delete_special)
        return diffs

    def _is_normal_value(self, key):
        """Return True IFF the value of selection `key` is not special, i.e. N/A or OMIT."""
        return not MappingSelectionsDict.is_special_value(self.selections[key])

    def _value_name(self, key):
        """Return either a special value,  or the filename of the loaded mapping."""
        value = self.selections[key]
        return value if MappingSelectionsDict.is_special_value(value) else value.filename

    def _value_xsum(self, key):
        """Return selections[key] if it is a special value, otherwise assume it is a mapping
        and return the sha1sum header field.
        """
        value = self.selections[key]
        return value if MappingSelectionsDict.is_special_value(value) else value.sha1sum

    def difference_header(self, other, path=(), pars=()):
        """Compare `self` with `other` and return a list of difference
        tuples,  prefixing each tuple with context `path`.
        """
        other = asmapping(other, cache="readonly")
        differences = []
        for key in self.header:
            if key not in other.header:
                diff = selectors.DiffTuple(
                    * path + ((self.filename, other.filename), "deleted header " + repr(key) + " = " + repr(self.header[key])),
                    parameter_names = pars + (self.diff_name, "DIFFERENCE",))
                differences.append(diff)
            elif self.header[key] != other.header[key]:
                diff = selectors.DiffTuple(
                    * path + ((self.filename, other.filename), "header replaced " + repr(key) + " = " + repr(self.header[key]) + " with " + repr(other.header[key])),
                    parameter_names = pars + (self.diff_name, "DIFFERENCE",))
                differences.append(diff)
        for key in other.header:
            if key not in self.header:
                diff = selectors.DiffTuple(
                    * path + ((self.filename, other.filename), "header added " + repr(key) + " = " + repr(other.header[key])),
                    parameter_names = pars + (self.diff_name, "DIFFERENCE",))
                differences.append(diff)
        return sorted(differences)

    @property
    def diff_name(self):
        """Name used to identify mapping item in DiffTuple's"""
        return self.__class__.__name__ # .replace("Context","").replace("Mapping","")

    def copy(self):
        """Return an in-memory copy of this rmap as a new object."""
        return self.from_string(self.format(), self.filename, ignore_checksum=True)

    def reference_names(self):
        """Returns set(ref_file_name...)"""
        return sorted({ reference for selector in self.selections.normal_values() for reference in selector.reference_names() })

    def reference_name_map(self):
        """Returns { filekind : set( ref_file_name... ) }"""
        name_map = { filekind:selector.reference_names() for (filekind, selector) in self.selections.normal_items() }
        name_map.update(dict(self.selections.special_items()))
        return name_map

    def mapping_names(self):
        """Returns a list of mapping files associated with this Mapping"""
        return sorted([self.basename] + [name for selector in self.selections.normal_values() for name in selector.mapping_names()])

    def file_matches(self, filename):
        """Return the "extended match tuples" which can be followed to arrive at `filename`."""
        return sorted([match for value in self.selections.normal_values() for match in value.file_matches(filename)])

    def get_derived_from(self):
        """Return the Mapping object `self` was derived from, or None."""
        for substring in self.null_derivation_substrings:
            if substring in self.derived_from:
                log.info("Skipping derivation checks for root mapping", repr(self.basename),
                         "derived_from =", repr(self.derived_from))
                return None
        derived_path = locate_mapping(self.derived_from)
        if os.path.exists(derived_path):
            with log.error_on_exception("Can't load parent mapping", repr(derived_path)):
                derived_from = fetch_mapping(derived_path)
                return derived_from
            return None
        else:
            log.warning("Parent mapping for", repr(self.basename), "=", repr(self.derived_from), "does not exist.")
            return None

    def _check_nested(self, key, self_val, nested):
        """Verify that `key` in `nested's` header matches `key` in `self's` header."""
        nested_val = getattr(nested, key)
        assert  self_val == nested_val, \
            "selector['{}']='{}' in '{}' doesn't match header['{}']='{}' in nested file '{}'.".format(
            key, self_val, self.filename, key, nested_val, nested.filename)

    def todict(self, recursive=10):
        """Return a 'pure data' dictionary representation of this mapping and it's children
        suitable for conversion to json.  If `recursive` is non-zero,  return that many
        levels of the hierarchy starting with this one.   If recursive is zero,  only
        return the filename and header of the next levels down,  not the contents.
        """
        # dict required here for subsequent update.
        selections = dict([(key, val.todict(recursive-1)) if recursive-1
                            else (val.basename, val.header)
                            for (key,val) in self.selections.normal_items()])
        selections.update(dict([(key, {"name": val, "text_descr": self.obs_package.TEXT_DESCR[key]})
                                for (key,val) in self.selections.special_items()]))
        # selections.items() critical below for prior interface,  web context display
        return {
                "header" : { key: self.header[key] for key in self.header },
                "name" : self.header["name"],
                "parameters" : tuple(self.parkey),
                "selections" : list(sorted(selections.items())),
                }

    def tojson(self, recursive=10):
        """Return a JSON representation of this mapping and it's children."""
        return json.dumps(self.todict(recursive=recursive))

    def get_instrument(self, header):
        """Return the name of the instrument which corresponds to `header`.   Called for unknown-mapping types.
        Overridden by PipelineMapping which figures it out from header.
        """
        return self.instrument.upper()

    def locate_file(self, filename):
        """Return the full path (in cache or absolute) of `filename` as determined by the current environment."""
        return locate_file(filename, self.observatory)

# ===================================================================

class ContextMapping(Mapping):
    """.pmap and .imap base class."""

    required_attrs = Mapping.required_attrs + ["observatory"]

    def __init__(self, filename, header, selector, **keys):
        super(ContextMapping, self).__init__(filename, header, selector, **keys)
        self.observatory = self.header["observatory"]
        self.selections = MappingSelectionsDict(selector, self.keys)
        if config.FORCE_COMPLETE_LOAD:
            self.force_load()

    def force_load(self):
        """Ensure that all submappings are loaded, i.e. make artificial demand."""
        for selection in self.selections.normal_values():
            selection.force_load()

    def set_item(self, key, value):
        """Add or replace and element of this mapping's selector.   For re-writing only.

        Return the case-corrected value of `key` and the value that was replaced or None
        as (corrected_key, replaced_value).
        """
        key, value = str(key), str(value)
        key = self.locate.match_context_key(key)
        replaced = self.selector.get(key, None)
        self.selector[key] = value
        return key, replaced

    def validate(self):
        """Validate nested mappings, common properties for all context mappings."""
        super(ContextMapping, self).validate()
        self.check_observatory()
        for key, mapping in self.selections.normal_items():
            self._check_nested("observatory", self.observatory, mapping)
            # mapping.validate()

# ===================================================================

class PipelineContext(ContextMapping):
    """A pipeline context describes the context mappings for each instrument
    of a pipeline.
    """
    mapping_type = "pipeline"

    def __init__(self, filename, header, selector, **keys):
        super(PipelineContext, self).__init__(filename, header, selector, **keys)
        self.instrument_key = self.parkey[0].upper()   # e.g. INSTRUME

    def validate(self):
        """Implement PipelineContext validations which require loading sub-mappings here."""
        super(PipelineContext, self).validate()
        for instrument, imap in self.selections.normal_items():
            self._check_nested("instrument", instrument, imap)

        # Will raise pkg_resources.RequirementParseError if the requirement
        # is malformed:
        self.get_asdf_standard_requirement()

    def get_best_references(self, header, include=None):
        """Return the best references for keyword map `header`.  If `include`
        is None,  collect all filekinds,  else only those listed.
        """
        header = dict(header)   # make a copy
        instrument = self.get_instrument(header)
        imap = self.get_imap(instrument)
        return imap.get_best_references(header, include)

    def get_old_references(self, header, include=None):
        """Return the old references defined in keyword map `header` using this
        context to define the types to return when `include` is None.

        >>> p = get_cached_mapping("hst.pmap")

        >>> h = {'CENTERA1': '513.0', 'CENTERA2': '513.0', 'PHOTCORR': 'UNDEFINED', 'FLSHCORR': 'UNDEFINED', 'MEMBER_NAME': 'I9ZF01DZQ', 'MEMBER_TYPE': 'EXP-DTH', 'SUBTYPE': 'FULLIMAG', 'OSCNTAB': 'Q911321MI_OSC.FITS', 'MEANEXP': '0.0', 'PFLTFILE': 'UC72113PI_PFL.FITS', 'CHINJECT': 'NONE', 'CCDCHIP': '0.0', 'EXPSTART': '2009-12-07 16:10:34.380000', 'FLATCORR': 'UNDEFINED', 'CRTHRESH': '0.0', 'CRRADIUS': '0.0', 'ASN_ID': 'I9ZF01010', 'INSTRUME': 'WFC3', 'SUBARRAY': 'F', 'CCDAMP': 'ABCD', 'CCDOFSTA': 'UNDEFINED', 'APERTURE': 'IR', 'SIZAXIS2': '1024.0', 'SIZAXIS1': '1024.0', 'FLSHFILE': 'N/A', 'BINAXIS1': '1.0', 'BINAXIS2': '1.0', 'CCDTAB': 'T2C16200I_CCD.FITS', 'IDCTAB': 'W3M18525I_IDC.FITS', 'SKYSUB': 'UNDEFINED', 'CCDOFSTB': 'UNDEFINED', 'CCDOFSTC': 'UNDEFINED', 'CCDOFSTD': 'UNDEFINED', 'DQICORR': 'UNDEFINED', 'FLASHCUR': 'UNDEFINED', 'DATA_SET': 'I9ZF01010', 'CRSIGMAS': 'UNDEFINED', 'DARKCORR': 'UNDEFINED', 'SAMP_SEQ': 'STEP200', 'SCALENSE': '0.0', 'DRIZCORR': 'UNDEFINED', 'INITGUES': 'UNDEFINED', 'IMPHTTAB': 'WBJ1825RI_IMP.FITS', 'SHADCORR': 'UNDEFINED', 'DATA_SET_EXP': 'I9ZF01DZQ', 'CRMASK': 'F', 'MDRIZTAB': 'UBI1853PI_MDZ.FITS', 'OBSMODE': 'MULTIACCUM', 'NLINFILE': 'U1K1727MI_LIN.FITS', 'ATODCORR': 'UNDEFINED', 'BADINPDQ': '0.0', 'SHUTRPOS': 'UNDEFINED', 'DATE-OBS': '2009-12-07', 'ATODTAB': 'N/A', 'TIME-OBS': '16:10:34.380000', 'DARKFILE': 'XAG19298I_DRK.FITS', 'FILTER': 'F110W', 'BPIXTAB': 'Y711519RI_BPX.FITS', 'CCDGAIN': '2.5', 'DETECTOR': 'IR', 'BIASFILE': 'N/A', 'CRSPLIT': 'UNDEFINED', 'CRREJTAB': 'U6A1748RI_CRR.FITS', 'BIASCORR': 'UNDEFINED'}

        >>> import pprint
        >>> pprint.pprint(p.get_old_references(h))
        {'ATODTAB': 'N/A',
         'BIASFILE': 'N/A',
         'BPIXTAB': 'Y711519RI_BPX.FITS',
         'CCDTAB': 'T2C16200I_CCD.FITS',
         'CRREJTAB': 'U6A1748RI_CRR.FITS',
         'DARKFILE': 'XAG19298I_DRK.FITS',
         'FLSHFILE': 'N/A',
         'IDCTAB': 'W3M18525I_IDC.FITS',
         'IMPHTTAB': 'WBJ1825RI_IMP.FITS',
         'MDRIZTAB': 'UBI1853PI_MDZ.FITS',
         'NLINFILE': 'U1K1727MI_LIN.FITS',
         'OSCNTAB': 'Q911321MI_OSC.FITS',
         'PFLTFILE': 'UC72113PI_PFL.FITS'}
        """
        header = dict(header)   # make a copy
        instrument = self.get_instrument(header)
        imap = self.get_imap(instrument)
        return imap.get_old_references(header, include)

    def get_imap(self, instrument):
        """Return the InstrumentMapping corresponding to `instrument`."""
        instrument_hacks = {
                "wfii" : "wfpc2",
            }
        instrument = instrument_hacks.get(instrument.lower(), instrument.lower())
        try:
            return self.selections[instrument]
        except (crexc.IrrelevantReferenceTypeError, crexc.OmitReferenceTypeError):
            raise
        except KeyError as exc:
            raise crexc.CrdsUnknownInstrumentError("Unknown instrument " + repr(instrument) +
                                  " for context " + repr(self.basename)) from exc

    def get_filekinds(self, dataset):
        """Return the filekinds associated with `dataset` by examining
        it's parameters.  Currently returns ALL filekinds for
        `dataset`s instrument.   Not all are necessarily appropriate for
        the current mode.  `dataset` can be a filename or a header dictionary.
        """
        if isinstance(dataset, str):
            from crds import data_file
            instrument = data_file.getval(dataset,  self.instrument_key)
        elif isinstance(dataset, dict):
            instrument = self.get_instrument(dataset)
        else:
            raise ValueError("Dataset should be a filename or header dictionary.")
        return self.get_imap(instrument).get_filekinds(dataset)

    def get_instrument(self, header):
        """Get the instrument name defined by file `header`."""
        for key in [self.instrument_key.upper(), self.instrument_key.lower()] + INSTRUMENT_KEYWORDS:
            try: # This hack makes FITS headers work prior to back-mapping to data model names.
                instr = header[key]
            except KeyError:
                continue
            else:
                if instr.upper() != "UNDEFINED":
                    break
        else:
            raise crexc.CrdsError("Missing '%s' keyword in header for determining instrument." % self.instrument_key)
        return instr.upper()

    def get_item_key(self, filename):
        """Given `filename` nominally to insert, return the instrument it corresponds to."""
        instrument, _filekind = utils.get_file_properties(self.observatory, filename)
        return instrument.upper()

    def get_equivalent_mapping(self, mapping):
        """Return the Mapping equivalent to name `mapping` in pmap `self`,  or None."""
        if mapping.endswith(".pmap"):
            return self
        else:
            instrument, _filekind = utils.get_file_properties(self.observatory, mapping)
            try:
                imap = self.get_imap(instrument)
            except Exception:
                log.warning("No equivalent instrument in", repr(self.name), "corresponding to", repr(mapping))
                return None
            else:
                return imap.get_equivalent_mapping(mapping)

    @utils.cached
    def get_required_parkeys(self):
        """Return a dictionary of matching parameters for each instrument:

            { instrument : [ matching_parkey_name, ... ], }
        """
        return { instrument : list(self.parkey) + self.selections[instrument].get_required_parkeys()
                 for instrument in self.selections.normal_keys() }

    def get_asdf_standard_requirement(self):
        raw_requirement = self.header.get("asdf_standard_requirement", None)
        if raw_requirement:
            raw_requirement = "asdf_standard " + raw_requirement
        else:
            raw_requirement = "asdf_standard"

        return Requirement.parse(raw_requirement)

# ===================================================================

class InstrumentContext(ContextMapping):
    """An instrument context describes the rmaps associated with each filetype
    of an instrument.
    """
    mapping_type = "instrument"

    required_attrs = ContextMapping.required_attrs + ["instrument"]

    def __init__(self, filename, header, selector, **keys):
        super(InstrumentContext, self).__init__(filename, header, selector, **keys)
        self.instrument = self.header["instrument"]

    def validate(self):
        """Perform InstrumentContext semantic checks which require can loading sub-mappings."""
        super(InstrumentContext, self).validate()
        self.check_instrument()
        for filekind, refmap in self.selections.normal_items():
            self._check_nested("instrument", self.instrument, refmap)
            self._check_nested("filekind", filekind, refmap)

    def get_rmap(self, filekind):
        """Given `filekind`,  return the corresponding ReferenceMapping."""
        filekind = str(filekind).lower()
        if filekind not in self.selections:
            raise crexc.CrdsUnknownReftypeError("Unknown reference type", repr(filekind))
        if MappingSelectionsDict.is_na_value(self.selections[filekind]):
            log.verbose("Reference type", repr(filekind),
                        "is declared N/A at the instrument level for", repr(self.instrument), verbosity=70)
            raise crexc.IrrelevantReferenceTypeError("Type", repr(filekind), "is N/A for", repr(self.instrument))
        if  MappingSelectionsDict.is_omit_value(self.selections[filekind]):
            log.verbose("Reference type", repr(filekind),
                        "is omitted at the instrument level for", repr(self.instrument), verbosity=70)
            raise crexc.OmitReferenceTypeError("Type", repr(filekind), "is OMITTED for", repr(self.instrument))
        return self.selections[filekind]

    def get_imap(self, instrument):
        """Make sequences like get_cached_mapping(mapping).get_imap(instrument)... work where `mapping` is an
        instrument mapping.
        """
        return self

    def get_best_references(self, header, include=None):
        """Returns a map of best references { filekind : reffile_basename }
        appropriate for this `header`.   If `include` is None, include all
        filekinds in the results,  otherwise compute and include only
        those filekinds listed.
        """
        refs = {}
        if not include:
            include = self.selections.keys()
        for filekind in include:
            log.verbose("-"*120, verbosity=55)
            filekind = filekind.lower()
            ref = None
            try:
                ref = self.get_rmap(filekind).get_best_ref(header)
            except crexc.IrrelevantReferenceTypeError:
                ref = "NOT FOUND n/a"
            except crexc.OmitReferenceTypeError:
                ref = None
            except Exception as exc:
                ref = "NOT FOUND " + str(exc)
            if ref is not None:
                refs[filekind] = ref
        log.verbose("-"*120, verbosity=55)
        return refs

    def get_old_references(self, header, include=None):
        """Returns a map of old references which were recorded in `header`,
        returning only those types listed in `include` or all types if
        `include` is None.
        """
        refs = {}
        if not include:
            include = self.selections.keys()
        include = [ key.upper() for key in include ]
        header = { key.upper() : val.strip() for (key, val) in header.items() }
        for filekind in include:
            try:
                ref = header[filekind]
                if ref == "NOT FOUND n/a":
                    ref = "N/A"
                refs[filekind] = ref
            except KeyError:
                refs[filekind] = "UNDEFINED"
        return refs

    def get_parkey_map(self):
        """Infers the legal values of each parkey from the rmap itself.
        This is a potentially different answer than that defined by the TPNs,
        the latter being considered definitive.
        Return { parkey : [legal values...], ... }
        """
        pkmap = {}
        for selection in self.selections.normal_values():
            for parkey, choices in selection.get_parkey_map().items():
                if parkey not in pkmap:
                    pkmap[parkey] = set()
                pkmap[parkey] |= set(choices)
        for parkey, choices in pkmap.items():
            pkmap[parkey] = list(pkmap[parkey])
            if "CORR" not in parkey:
                pkmap[parkey].sort()
        return pkmap

    def get_valid_values_map(self, condition=False, remove_special=True):
        """Based on the TPNs,  return a mapping from parkeys to their valid
        values for all parkeys for all filekinds of this instrument.   This will
        return the definitive lists of legal values,  not all of which are
        required to be represented in rmaps;  these are the values that *could*
        be in an rmap,  not necessarily what is in any given rmap to match.

        If `condition` is True,  values are filtered with
        utils.condition_value() to match their rmap string appearance.   If
        False,  values are returned as raw TPN values and types.

        If `remove_special` is True,  values of ANY or N/A are removed from the
        lists of valid values.
        """
        pkmap = {}
        for selection in self.selections.normal_values():
            rmap_pkmap = selection.get_valid_values_map(condition)
            for key in rmap_pkmap:
                if key not in pkmap:
                    pkmap[key] = set()
                pkmap[key] |= set(rmap_pkmap[key])
        for key in self.get_parkey_map():
            if key not in pkmap:
                pkmap[key] = []    # flag a need for an unconstrained input
        if remove_special:
            specials = {"ANY","N/A"}
            for key in pkmap:  # remove specials like ANY or N/A
                if pkmap[key]:
                    pkmap[key] = pkmap[key] - specials
        for key in pkmap:  # convert to sorted lists
            pkmap[key] = sorted(pkmap[key])
        return pkmap

    def get_filekinds(self, dataset=None):
        """Return the filekinds associated with this dataset,  ideally
        the minimum set associated with `dataset`,  but initially all
        for dataset's instrument,  assumed to be self.instrument.
        """
        return list(sorted(self.selections.keys()))

    def get_item_key(self, filename):
        """Given `filename` nominally to insert, return the filekind it corresponds to."""
        _instrument, filekind = utils.get_file_properties(self.observatory, filename)
        return filekind.upper() if self.observatory == "jwst" else filekind.lower()

    def get_equivalent_mapping(self, mapping):
        """Return the Mapping equivalent to name `mapping` in imap `self`, or None."""
        if mapping.endswith(".pmap"):
            log.warning("Invalid comparison context", repr(self.name), "for", repr(mapping))
            return None
        if mapping.endswith(".imap"):
            return self
        else:
            _instrument, filekind = utils.get_file_properties(self.observatory, mapping)
            try:
                rmap = self.get_rmap(filekind)
            except Exception:
                log.warning("No equivalent filekind in", repr(self.name), "corresponding to", repr(mapping))
                return None
            else:  # I think it's always just "rmap".
                return rmap.get_equivalent_mapping(mapping)

    def difference(self, *args, **keys):
        """difference specialized to add .instrument to diff."""
        diffs = super(InstrumentContext, self).difference(*args, **keys)
        for diff in diffs:
            diff.instrument = self.instrument
        return diffs

# ===================================================================

class ReferenceMapping(Mapping):
    """ReferenceMapping manages loading the rmap associated with a single
    reference filetype and instantiate an appropriate selector tree from the
    rmap header and data.
    """
    mapping_type = "reference"
    required_attrs = Mapping.required_attrs + ["observatory","instrument","filekind"]
    specifies_references = True

    def __init__(self, *args, **keys):
        super(ReferenceMapping, self).__init__(*args, **keys)
        self.observatory = self.header["observatory"]
        self.instrument = self.header["instrument"]
        self.filekind = self.header["filekind"]

        self.schema_uri = self.header.get("schema_uri", None)

        self._reffile_switch = self.header.get("reffile_switch", "NONE").upper()
        self._reffile_format = self.header.get("reffile_format", "IMAGE").upper()
        self._reffile_required = self.header.get("reffile_required", "NONE").upper()

        # header precondition method, e.g. crds.hst.acs.precondition_header
        # TPNs define the static definitive possibilities for parameter choices
        # rmaps define the actually appearing literal parameter values
        self._rmap_valid_values = self.selector.get_value_map()
        self._required_parkeys = self.get_required_parkeys()

        # For "rmap_relevance" and "rmap_omit" expressions,  the expressions are enclosed in ()
        # to ensure no case conversions occur in LowerCaseDict.  Since ALWAYS is not in (),  it
        # shows up here as the standard "always".

        # if _rmap_relevance_expr evaluates to True at match-time,  this is a relevant type for that header.
        self._rmap_relevance_expr = None

        # if _rmap_omit_expr evaluates to True at match-time,  this type should be omitted from bestrefs results.
        self._rmap_omit_expr = None

        # for each parkey in parkey_relevance_exprs,  if the expr evaluates False,  it is mapped to N/A at match time.
        self._parkey_relevance_exprs = None

        # header precondition method, e.g. crds.hst.acs.precondition_header
        # this is optional code which pre-processes and mutates header inputs
        # set to identity if not defined.
        self._precondition_header = None

        # function to mutate to secondary lookup header if first lookup fails,  if defined.
        self._fallback_header = None

        # rmap-update-time hook to generate reference header variations for more complex but otherwise static lookups
        # this is a potential alternative to preconditioning and fallback which are *runtime* hooks and non-transparent
        # to someone looking at the rmap.
        self._rmap_update_headers = None

        # Actually compile lambdas for the hooks above.
        self._init_compiled()

    def __getstate__(self):
        """Return rmap pickling state,  minus lambdas and anything else that doesn't pickle."""
        state = dict(self.__dict__)
        del state["_rmap_relevance_expr"]
        del state["_rmap_omit_expr"]
        del state["_parkey_relevance_exprs"]
        del state["_precondition_header"]
        del state["_fallback_header"]
        del state["_rmap_update_headers"]
        return state

    def __setstate__(self, state):
        """Recreate rmap object from `state`,  recompiling missing __getstate__ objects on the fly."""
        self.__dict__ = dict(state)
        self._init_compiled()

    def force_load(self):
        """Nothing below ReferenceMapping is loaded."""
        pass

    def _init_compiled(self):
        """Initialize object fields which contain compiled code objects, special handling for pickling."""
        self._comment_parkeys = tuple(name.lower() for name in self.header.get("comment_parkeys", ()))
        self._rmap_relevance_expr = self.get_expr(self.header.get("rmap_relevance", "always").replace("always", "True"))
        self._rmap_omit_expr = self.get_expr(self.header.get("rmap_omit", "False"))

        relevant  = dict(self.header.get("parkey_relevance", {}))
        relevant.update({
            name : "keep_comments" for name in self._comment_parkeys
        })
        self._parkey_relevance_exprs = {
            name.lower() : self.get_expr(expr) for (name, expr) in relevant.items()
            }

        self._precondition_header = self.get_hook("precondition_header", (lambda self, header: header))
        self._fallback_header = self.get_hook("fallback_header", (lambda self, header: None))
        self._rmap_update_headers = self.get_hook("rmap_update_headers", None)

    def validate(self):
        """Validate the contents of this rmap against the TPN for this
        filekind / reftype.   Each field of each Match tuple must have a value
        OK'ed by the TPN.  UseAfter dates must be correctly formatted.
        """
        super(ReferenceMapping, self).validate()
        self.check_observatory()
        self.check_instrument()
        self.check_filekind()
        self.check_schema_uri()
        if "reference_to_dataset" in self.header:
            parkeys = self.get_required_parkeys()
            for _reference, dataset in self.reference_to_dataset.items():
                assert dataset.upper() in parkeys, \
                    "reference_to_dataset dataset keyword not in parkey keywords."
        with log.augment_exception("Invalid mapping:", self.instrument, self.filekind):
            self.selector.validate_selector(self.tpn_valid_values)

    def check_schema_uri(self):
        """
        Check that the schema URI corresponds to a schema that
        ASDF knows how to load.
        """
        import asdf

        if self.schema_uri is not None:
            with log.augment_exception("Invalid ASDF schema URI:", self.schema_uri):
                asdf.schema.load_schema(self.schema_uri)

    def get_expr(self, expr):  # secured
        """Return (expr, compiled_expr) for some rmap header expression, generally a predicate which is evaluated
        in the context of the matching header to fine tune behavior.   Screen the expr for dangerous code.
        """
        expr = utils.condition_source_code_keys(expr, self.get_required_parkeys())
        try:
            return expr, MAPPING_VERIFIER.compile_and_check(expr, source=self.basename, mode="eval")
        except crexc.MappingFormatError as exc:
            raise crexc.MappingFormatError("Can't load file " + repr(self.basename) + " : " + str(exc)) from exc

    def get_hook(self, name, default):
        """Return plugin hook function generically named `name` or `default` if `name` is not defined in
        the associated instrument package or in the rmap header.   Until hooks is defined in header,  get_hook
        will return the generically named function in the instrument package.   hooks in rmap header supports
        future versions of hooks without breaking past rmaps.

        Hooks are called basically like methods, typically as hook(self, header).   Hooks return hook-specific values.

        Nth generation hooks defined by name in rmap header "hooks" { unversioned_name : versioned_hook_name } dict.
        To unplug a hook,  define it in rmap header "hooks" like { "precondition_header" : "none" }
        """
        hooks = self.header.get("hooks", None)
        if hooks:  # Either get the replacement name,  or use the original name if not found.
            hook_name = hooks.get(name, name)
        else:  # No hooks dict,  just use the original name.
            hook_name = "_".join([name, self.instrument, self.filekind, "v1"])
        hook = getattr(self.instr_package, hook_name, default)
        if hook is not default:
            log.verbose("Using hook", repr(hook_name), "for rmap", repr(self.basename), verbosity=55)
        return hook

    # Unusual caching style implements deferred loading of .tpn files,  fairly slow.
    @property
    @utils.cached
    def tpn_valid_values(self):
        """Property, dictionary of valid values for each parameter loaded from .tpn files or equivalent."""
        return self.get_valid_values_map()

    def get_filekinds(self, dataset=None):
        """Return the filekinds associated with this dataset, but for an rmap,
        just the rmap's filekind.
        """
        return [self.filekind]

    def get_best_references(self, header, include=None):
        """Shim so that .rmaps can be used for bestrefs in place of a .pmap or .imap for single type development."""
        if include is not None and self.filekind not in include:
            raise crexc.CrdsUnknownReftypeError(self.__class__.__name__, repr(self.basename),
                                          "can only compute bestrefs for type", repr(self.filekind), "not", include)
        bestref = self.get_best_ref(header)
        if bestref is not None:
            return { self.filekind : self.get_best_ref(header) }
        else:
            return {}

    def get_best_ref(self, header):
        """Return a single best reference value associated with this .rmap and `header`.  Map exceptions
        from nested methods onto simple "NOT FOUND..." strings which are exempted from reference downloads.
        """
        try:
            return self._get_best_ref(header)
        except crexc.IrrelevantReferenceTypeError:
            return "NOT FOUND n/a"
        except crexc.OmitReferenceTypeError:
            return None
        except Exception as exc:
            if log.get_exception_trap():
                return "NOT FOUND " + str(exc)
            else:
                raise

    def _get_best_ref(self, header_in):
        """Return the single reference file basename appropriate for
        `header_in` selected by this ReferenceMapping.
        """
        header_in = dict(header_in)
        log.verbose("Getting bestrefs:", self.basename, verbosity=55)
        expr_header = utils.condition_header_keys(header_in)
        self.check_rmap_omit(expr_header)     # Should bestref be omitted based on rmap_omit expr?
        self.check_rmap_relevance(expr_header)  # Should bestref be set N/A based on rmap_relevance expr?
        # Some filekinds, .e.g. ACS biasfile, mutate the header
        header = self._precondition_header(self, header_in) # Execute type-specific plugin if applicable
        header = self.map_irrelevant_parkeys_to_na(header)  # Execute rmap parkey_relevance conditions
        try:
            bestref = self.selector.choose(header)
        except Exception as exc:
            log.verbose("First selection failed:", str(exc), verbosity=55)
            header = self._fallback_header(self, header_in) # Execute type-specific plugin if applicable
            try:
                if header:
                    header = self.minimize_header(header)
                    log.verbose("Fallback lookup on", repr(header), verbosity=55)
                    header = self.map_irrelevant_parkeys_to_na(header) # Execute rmap parkey_relevance conditions
                    bestref = self.selector.choose(header)
                else:
                    raise
            except Exception as exc:
                log.verbose("Fallback selection failed:", str(exc), verbosity=55)
                if self._reffile_required in ["YES", "NONE"]:
                    log.verbose("No match found and reference is required:",  str(exc), verbosity=55)
                    raise
                else:
                    log.verbose("No match found but reference is not required:",  str(exc), verbosity=55)
                    raise crexc.IrrelevantReferenceTypeError("No match found and reference type is not required.") from exc
        log.verbose("Found bestref", repr(self.instrument), repr(self.filekind), "=", repr(bestref), verbosity=55)
        if MappingSelectionsDict.is_na_value(bestref):
            raise crexc.IrrelevantReferenceTypeError("Rules define this type as Not Applicable for these observation parameters.")
        if MappingSelectionsDict.is_omit_value(bestref):
            raise crexc.OmitReferenceTypeError("Rules define this type to be Omitted for these observation parameters.")
        return bestref

    def reference_names(self):
        """Return the list of reference file basenames associated with this
        ReferenceMapping.
        """
        return [name for name in self.selector.reference_names() if not is_special_value(name)]

    def mapping_names(self):
        """Return name of this ReferenceMapping as degenerate list of 1 item."""
        return [self.basename]

    @utils.cached
    def get_required_parkeys(self, include_reffile_switch=True):
        """Return the list of parkey names needed to select from this rmap."""
        parkeys = []
        for key in self.parkey:
            if isinstance(key, tuple):
                parkeys += list(key)
            else:
                parkeys.append(key)
        if include_reffile_switch and self._reffile_switch != "NONE":
            parkeys.append(self._reffile_switch)
        parkeys.extend(list(self.extra_keys))
        parkeys.extend(getattr(self.obs_package, "extra_keys", []))
        return parkeys

    def get_extra_parkeys(self):
        """Return a tuple of parkeys which are not directly matched.   These correspond
        to HST dataset parkeys which were used to compute the values of other keys
        which *are* used to match.   These keys appear in HST rmaps with constant
        universal values of N/A.  At rmap update time,  these keys need to be mapped
        to N/A in the event they're actually defined in the reference to avoid creating
        new rules for that specific case when the parameter is not really intended for matching.
        """
        extra = list(self.extra_keys)
        if self._reffile_switch != "NONE":
            extra.append(self._reffile_switch)
        return tuple(extra)

    def get_parkey_map(self):
        """Based on the rmap,  return the mapping from parkeys to their
        handled values,  i.e. what this rmap says it matches against.
        Note that these are the values seen in the rmap prior to any
        substitutions which are defined in the header.

        Return { parkey : [match values, ...], ... }
        """
        parkey_map = self.selector.get_parkey_map()
        tpn_values = self.tpn_valid_values
        for key in self.get_extra_parkeys():
            if key in parkey_map and "CORR" not in key:
                continue
            parkey_map[key] = tpn_values.get(key, [])
            if key.endswith("CORR"):  #  and parkey_map[key] == []:
                parkey_map[key] = ["PERFORM", "OMIT", "NONE", "COMPLETE", "UNDEFINED"]
        return parkey_map

    def get_valid_values_map(self, condition=True):
        """Based on the TPNs,  return a mapping from each of the required
        parkeys to its valid values,

        i.e. the definitive source for what is legal for this filekind.

        return { parkey : [ valid values ] }
        """
        tpninfos = self.locate.get_all_tpninfos(self.instrument, self.filekind, "ld_tpn")
        required_keys = self.get_required_parkeys()
        valid_values = {}
        for info in tpninfos:
            if info.is_complex_constraint:
                continue
            if info.name in required_keys:
                values = info.values
                if len(values) == 1 and ":" in values[0]:
                    limits = values[0].split(":")
                    try:
                        limits = [int(float(x)) for x in limits]
                    except Exception:
                        pass
                        # sys.exc_clear()
                    else:
                        values = list(range(limits[0], limits[1]+1))
                if condition:
                    values = tuple([utils.condition_value(val) for val in values])
                valid_values[info.name] = values
        return valid_values

    def file_matches(self, filename):
        """Return a list of the match tuples which refer to `filename`."""
        sofar = ((("observatory", self.observatory),
                  ("instrument",self.instrument),
                  ("filekind", self.filekind),),)
        return sorted(self.selector.file_matches(filename, sofar))

    def difference(self, other, path=(), pars=(), include_header_diffs=False, recurse_added_deleted=False):
        """Return the list of difference tuples between `self` and `other`, prefixing each tuple with context `path`.
        Elements of `path` are named by correspnding elements of `pars`.
        """
        other = asmapping(other, cache="readonly")
        header_diffs = self.difference_header(other, path=path, pars=pars) if include_header_diffs else []
        body_diffs = self.selector.difference(other.selector,
                path = path + ((self.filename, other.filename),),
                pars = pars + (self.diff_name,))
        diffs = header_diffs + body_diffs
        for diff in diffs:
            diff.instrument = self.instrument
            diff.filekind = self.filekind
        return diffs

    def diff_files(self, added_deleted, path=(), pars=()):
        """Return the list of diff tuples for all nested changed files in a higher level addition
        or deletion.   added_deleted should be "added" or "deleted"
        """
        body_diffs = self.selector.flat_diff(added_deleted + " terminal",
                path = path + ((self.filename,)), pars = pars + (self.diff_name,))
        for diff in body_diffs:
            diff.instrument = self.instrument
            diff.filekind = self.filekind
        return body_diffs

    def check_rmap_relevance(self, header):
        """Raise an exception if this rmap's relevance expression evaluated in the context of `header` returns False.
        """
        try:
            source, compiled = self._rmap_relevance_expr
            relevant = eval(compiled, {}, header)   # secured
            log.verbose("Filekind ", repr(self.instrument), repr(self.filekind),
                        "is relevant:", relevant, repr(source), verbosity=55)
        except Exception as exc:
            log.warning("Failed checking relevance for", repr(self.instrument),
                        repr(self.filekind), "with expr", repr(source),
                        ":", str(exc))
        else:
            if not relevant:
                raise crexc.IrrelevantReferenceTypeError(
                    "Rmap does not apply to the given parameter set based on rmap_relevance expression.")

    def check_rmap_omit(self, header):
        """Return True IFF this type should be omitted based on the 'rmap_omit' header expression."""
        source, compiled = self._rmap_omit_expr
        try:
            omit = eval(compiled, {}, header)   # secured
            log.verbose("Filekind ", repr(self.instrument), repr(self.filekind),
                        "should be omitted: ", omit, repr(source), verbosity=55)
        except Exception as exc:
            log.warning("Failed checking OMIT for", repr(self.instrument),
                        repr(self.filekind), "with expr", repr(source),
                        ":", str(exc))
        else:
            if omit:
                raise crexc.OmitReferenceTypeError("rmap_omit expression indicates this type should be omitted.")

    def map_irrelevant_parkeys_to_na(self, header, keep_comments=False):
        """Evaluate any relevance expression for each parkey, and if it's
        false,  then change the value to N/A.   Changing a parkey to N/A effectively
        removes it as a consideration for bestrefs.   Changing a parkey to N/A during
        rmap updates facilitates exact case matches (replacements) ignoring irrelevant
        parameters while defining the case.   If `keep_comments` is set to True then
        irrelevant_parkeys which are also comments are not converted to N/A.

        COME FROM: This function is called both on dataset headers during bestrefs and
        on reference file headers during rmap updates with the presumption that any
        parameter required by the relevance expressions is defined in both datasets and
        reference files.
        """
        from crds import data_file
        expr_header = dict(header)
        expr_header = data_file.ensure_keys_defined(expr_header, needed_keys=self._required_parkeys)
        expr_header = utils.condition_header_keys(expr_header)
        header = dict(header)  # copy
        expr_header["keep_comments"] = keep_comments
        for parkey in self._required_parkeys:  # Only add/overwrite irrelevant
            lparkey = parkey.lower()
            if lparkey in self._parkey_relevance_exprs:
                source, compiled = self._parkey_relevance_exprs[lparkey]
                relevant = eval(compiled, {}, expr_header)  # secured
                log.verbose("Parkey", self.instrument, self.filekind, lparkey,
                            "is relevant:", relevant, repr(source), verbosity=55)
                if not relevant:
                    log.verbose("Setting irrelevant parkey", repr(parkey), "to N/A")
                    header[parkey] = "N/A"
        return header

    def insert_reference(self, reffile):
        """Returns new ReferenceMapping made from `self` inserting `reffile`."""
        header = self.get_refactor_header(reffile)
        return self.insert_header_reference(header, os.path.basename(reffile))

    def get_refactor_header(self, reffile, extra_keys=()):
        """Give reference path `reffile` return the header which should be used to insert
        the file into this rmap based on the reffile contents.

        1. Loads basic header from reffile.
        2. Ensures required parkeys are at least defined as UNDEFINED or N/A.
        3. Maps from reference file keywords and values to dataset keywords and values rmaps match on.
        """
        # Since expansion rules may depend on keys not used in matching,  get entire header
        from crds import data_file
        header = data_file.get_header(reffile, observatory=self.observatory)
        needed_keys = tuple(self.get_reference_parkeys()) + tuple(extra_keys)
        header = data_file.ensure_keys_defined(header, needed_keys=needed_keys)
        # NOTE: required parkeys are in terms of *dataset* headers,  not reference headers.
        log.verbose("insert_reference raw reffile header:\n",
                    log.PP([ (key,val) for (key,val) in header.items() if key in self.get_reference_parkeys() ]),
                    verbosity=70)
        header = self.reference_to_dataset_header(header)
        log.verbose("insert_reference transformed-to-dataset header:\n",
                    log.PP([ (key,val) for (key,val) in header.items() if key in self.get_reference_parkeys() ]),
                    verbosity=70)
        return header

    def reference_to_dataset_header(self, header):
        """Convert the applicable keys in `header` from how they appear in the reference
        file to how they appear in the rmap. Multiple conversion activities:

        1. Mapping from reference file keywwords to corresponding dataset keywords.
        2. Expansion of conditional substitutions defined in e.g. crds.hst.substitutions.dat
        3. Any header conditioning,  normalization of values.
        4. Irrelevant parkeys are mapped to N/A if discernable based on `header`.
        5. extra_parkey keys are mapped to N/A.
        """
        # The reference file key and dataset key matched aren't always the same!
        # Specifically ACS BIASFILE NUMCOLS,NUMROWS and NAXIS1,NAXIS2
        # Also, DATE-OBS, TIME-OBS  <-->  USEAFTER
        header = self.locate.reference_keys_to_dataset_keys(self, header)

        # Reference files specify things like ANY which must be expanded to
        # glob patterns for matching with the reference file.
        header = substitutions.expand_wildcards(self, header)

        # Translate header values to .rmap normalized form,  e.g. utils.condition_value()
        header = self.locate.condition_matching_header(self, header)

        # Evaluate parkey relevance rules in the context of header to map
        # mode irrelevant parameters to N/A.
        # XXX not clear if/how this works with expanded wildcard or-patterns.
        header = self.map_irrelevant_parkeys_to_na(header, keep_comments=True)

        # The "extra" parkeys always appear in the rmap with values of "N/A".
        # The dataset value of the parkey is typically used to compute other parkeys
        # for HST corner cases.   It's a little stupid for them to appear in the
        # rmap match tuples,  but the dataset values for those parkeys are indeed
        # relevant,  and it does provide a hint that magic is going on.  At rmap update
        # time,  these parkeys need to be set to N/A even if they're actually defined.
        for key in self.get_extra_parkeys():
            log.verbose("Mapping extra parkey", repr(key), "from", header[key], "to 'N/A'.")
            header[key] = "N/A"
        return header

    def insert_header_reference(self, header, reffile):
        """Returns new ReferenceMapping made from `self` inserting `reffile`."""
        if self._rmap_update_headers:
            # Generate variations on header as needed to emulate header "pre-conditioning" and fall back scenarios.
            for hdr in self._rmap_update_headers(self, header):
                new = self.insert(hdr, reffile)
        else:
            # almost all instruments/types do this.
            new = self.insert(header, reffile)
        return new

    def get_reference_parkeys(self):
        """Return parkey names from the reference file perspective,  this can be a superset
        of the obvious parkey count due to redundant keyword expressions resulting from
        reference_to_dataset and mixed filetypes (some translated,  some not.)
        """
        dataset_parkeys = self.get_required_parkeys()
        if "reference_to_dataset" in self.header:
            for reference, dataset in self.reference_to_dataset.items():
                if dataset in dataset_parkeys:
                    dataset_parkeys.append(reference)
        return tuple(sorted(set(dataset_parkeys)))

    def insert(self, header, value):
        """Given reference file `header` and terminal `value`, insert the value into a copy
        of this rmap and return it.
        """
        new = self.copy()
        new.selector.insert(header, value,
            self.tpn_valid_values if not config.ALLOW_BAD_PARKEY_VALUES else {})
        return new

    def delete(self, terminal):
        """Remove all instances of `terminal` (nominally a filename) from `self`."""
        new = self.copy()
        terminal = os.path.basename(terminal)
        deleted_count = new.selector.delete(terminal)
        if deleted_count == 0:
            raise crexc.CrdsError("Terminal '%s' could not be found and deleted." % terminal)
        return new

    def todict(self, recursive=10):
        """Return a 'pure data' dictionary representation of this mapping and it's children
        suitable for conversion to json.  `recursive` is ignored.
        """
        nested = self.selector.todict_flat()
        return {
                "header" : { key : self.header[key] for key in self.header },
                "text_descr" : self.obs_package.TEXT_DESCR[self.filekind],
                "parameters" : tuple(nested["parameters"]),
                "selections" : nested["selections"]
                }

    def get_equivalent_mapping(self, mapping):
        """Return `self` for comparison if `mapping` name specifies an rmap, otherwise None."""
        if not mapping.endswith(".rmap"):
            log.warning("Invalid comparison context", repr(self.name), "for", repr(mapping))
            return None
        return self

    def _trace_compare(self, other, show_equal=False):
        utils.trace_compare(self, other, show_equal)

# ===================================================================

def _load(mapping, **keys):
    """Stand-off function to call load_mapping, fetch_mapping, or get_cached_mapping
    depending on the "loader" value of `keys`.
    """
    return keys["loader"](mapping, **keys)

def get_cached_mapping(mapping, **keys):
    """Load `mapping` from the file system or cache,  adding it and all it's
    descendents to the cache.

    NOTE:   mutations to the mapping are reflected in the cache.   This call is
    not suitable for experimental mappings which need to be reloaded from the
    file system since the cached version will be returned instead.   This call
    always returns the same Mapping object for a given set of parameters so it
    should not be used where a copy is required.

    Return a PipelineContext, InstrumentContext, or ReferenceMapping.
    """
    keys["loader"] = get_cached_mapping
    return _load_mapping(mapping, **keys)

def fetch_mapping(mapping, **keys):
    """Load any `mapping`,  exploiting Mapping's already in the cache but not
    adding anything extra.   This is safe for experimental mappings and temporaries
    because new mappings not in the cache are not added to the cache.

    This call only returns a copy of mappings not already in the cache.

    Return a PipelineContext, InstrumentContext, or ReferenceMapping.
    """
    keys["loader"] = fetch_mapping
    return _load_mapping.readonly(mapping, **keys)

def load_mapping(mapping, **keys):
    """Load any `mapping`,  ignoring the cache.   Returns a unique object
    for each call.   Slow but safe for any use,  reads every file and
    returns a new copy.

    Return a PipelineContext, InstrumentContext, or ReferenceMapping.
    """
    keys["loader"] = load_mapping
    return _load_mapping.uncached(mapping, **keys)

@utils.xcached(omit_from_key=["loader", "ignore_checksum"])
def _load_mapping(mapping, **keys):
    """_load_mapping fetches `mapping` from the file system or cache."""
    if mapping.endswith(".pmap"):
        cls = PipelineContext
    elif mapping.endswith(".imap"):
        cls = InstrumentContext
    elif mapping.endswith(".rmap"):
        cls = ReferenceMapping
    else:
        if config.S3_ENABLED:
            m = Mapping.from_s3(mapping, **keys)
        else:
            m = Mapping.from_file(mapping, **keys)
        mapping_type = m.header["mapping"].lower()
        if  mapping_type == "pipeline":
            cls = PipelineContext
        elif mapping_type == "instrument":
            cls = InstrumentContext
        elif mapping_type == "reference":
            cls = ReferenceMapping
        else:
            raise ValueError("Unknown mapping type for " + repr(mapping))

    if config.S3_ENABLED:
        return cls.from_s3(mapping, **keys)
    else:
        return cls.from_file(mapping, **keys)

def asmapping(filename_or_mapping, cached=False, **keys):
    """Return the Mapping object corresponding to `filename_or_mapping`.
    filename_or_mapping must either be a string (filename to be loaded) or
    a Mapping subclass which is simply returned.

    cached can be set to:

    False, "uncached"   ignore the mappings cache,  always reload,  don't add to cache
    "readonly"          load the mapping from the cache if possible,  don't add to the cache if not present
    True, "cached"      load the mapping from the cache if possible, add it to the cache if not present

    'readonly' is for experimental/proposed mappings which should not permanently exist in the
    cache because they may later be rejected and/or redefined.
    """
    if isinstance(filename_or_mapping, Mapping):
        return filename_or_mapping
    elif isinstance(filename_or_mapping, str):
        if cached in [False, "uncached"]:
            return load_mapping(filename_or_mapping, **keys)
        elif cached in [True, "cached"]:
            return get_cached_mapping(filename_or_mapping, **keys)
        elif cached == "readonly":
            return fetch_mapping(filename_or_mapping, **keys)
        else:
            raise ValueError("asmapping: cached must be in [True, 'cached', False, 'uncached','readonly']")
    else:
        raise TypeError("asmapping: parameter should be a string or mapping.")

# =============================================================================

class MappingSelectionsDict(LazyFileDict):
    """MappingSelectionsDict is a LazyFileDict with customized special values specific to CRDS.
    Mappings.

    >>> selections = MappingSelectionsDict({"this" : "N/A", "that":"hst_acs.imap"})

    >>> selections.normal_keys()
    ['that']

    >>> selections.special_keys()
    ['this']

    >>> selections.normal_values()
    [InstrumentContext('hst_acs.imap')]

    >>> selections.special_values()
    ['N/A']

    >>> list(selections.normal_items())
    [('that', InstrumentContext('hst_acs.imap'))]

    >>> list(selections.special_items())
    [('this', 'N/A')]

    """
    na_values_set = { "N/A", "TEMP_N/A", "n/a", "temp_n/a"}
    omit_values_set = { "OMIT", "TEMP_OMIT", "omit", "temp_n/a"}
    special_values_set = na_values_set | omit_values_set

    test_descr = {
        "N/A" : "Not Applicable",
        "TEMP_N/A" : "Temporarily Not Applicable",
        }

    def __init__(self, selector, load_keys={}):
        if "loader" not in load_keys:
            load_keys["loader"] = load_mapping
        super(MappingSelectionsDict, self).__init__(selector, load_keys)

    def transform_key(self, key):
        """Perform key lookups uniformly in lower case regardless of input."""
        return key.lower()

    @classmethod
    def is_na_value(cls, value):
        return isinstance(value, str) and value in cls.na_values_set

    @classmethod
    def is_omit_value(cls, value):
        return isinstance(value, str) and value in cls.omit_values_set

# =============================================================================

"""glob_pattern may not identify file type,  hence discrete versions."""

def list_references(glob_pattern, observatory, full_path=False):
    """Return the list of cached references for `observatory` which match `glob_pattern`."""
    references = []
    for path in utils.get_reference_paths(observatory):
        pattern = os.path.join(path, glob_pattern)
        references.extend(_glob_list(pattern, full_path))
    if full_path:
        references = [ref for ref in references if not os.path.isdir(ref)]
    return sorted(set(references))

def list_mappings(glob_pattern, observatory, full_path=False):
    """Return the list of cached mappings for `observatory` which match `glob_pattern`."""
    pattern = config.locate_mapping(glob_pattern, observatory)
    mappings = _glob_list(pattern, full_path)
    if full_path:
        mappings = [mapping for mapping in mappings if not os.path.isdir(mapping)]
    return sorted(set(mappings))

def list_pickles(glob_pattern, observatory, full_path=False):
    """Return the list of cached mappings for `observatory` which match `glob_pattern`."""
    pattern = config.locate_pickle(glob_pattern, observatory)
    pickles = _glob_list(pattern, full_path)
    if full_path:
        pickles = [pkl for pkl in pickles if not os.path.isdir(pkl)]
    return sorted(set(pickles))

def _glob_list(pattern, full_path=False):
    """Return the sorted glob of `pattern`, with/without path depending on `full_path`."""
    if full_path:
        return sorted(glob.glob(pattern))
    else:
        return sorted([os.path.basename(fpath) for fpath in glob.glob(pattern)])

# =============================================================================

def mapping_type(mapping):
    """
    >>> mapping_type("hst.pmap")
    'pmap'
    >>> mapping_type("hst_acs.imap")
    'imap'
    >>> mapping_type("hst_acs_biasfile.rmap")
    'rmap'
    >>> try:
    ...    mapping_type("hst_acs.foo")
    ... except IOError:
    ...    pass
    >>> mapping_type(get_cached_mapping('hst.pmap'))
    'pmap'
    >>> mapping_type(get_cached_mapping('hst_acs.imap'))
    'imap'
    >>> mapping_type(get_cached_mapping('hst_acs_darkfile.rmap'))
    'rmap'
    """
    if isinstance(mapping, str):
        if config.is_mapping(mapping):
            return os.path.splitext(mapping)[1][1:]
        else:
            mapping = fetch_mapping(mapping, ignore_checksum=True)
    if isinstance(mapping, PipelineContext):
        return "pmap"
    elif isinstance(mapping, InstrumentContext):
        return "imap"
    elif isinstance(mapping, ReferenceMapping):
        return "rmap"
    else:
        raise ValueError("Unknown mapping type for " + repr(Mapping))

# ===================================================================

def is_special_value(filename):
    """Return True IFF `filename` is one of the special values which
    does not correspond to a file, e.g. N/A.

    >>> is_special_value("foo.fits")
    False

    >>> is_special_value("N/A")
    True
    """
    return MappingSelectionsDict.is_special_value(str(filename))

# ============================================================================

def replace_rmap_text(rmapping, new_filename, old_text, new_text, *args, **keys):
    """Do simple text replacement from `old_text` to `new_text` in `rmapping`.
    """
    log.info("Replacing", srepr(old_text), "with", srepr(new_text), "in",
             srepr(rmapping.basename), "to", srepr(new_filename))
    original_rmap = str(rmapping)
    new_rmap = original_rmap.replace(old_text, new_text)
    new_mapping = ReferenceMapping.from_string(new_rmap, ignore_checksum=True)
    new_mapping.write(new_filename)

# ===================================================================

def get_best_references(context_file, header, include=None, condition=True):
    """Compute the best references for `header` for the given CRDS
    `context_file`.   This is a local computation using local rmaps and
    CPU resources.   If `include` is None,  return results for all
    filekinds appropriate to `header`,  otherwise return only those
    filekinds listed in `include`.
    """
    ctx = asmapping(context_file, cached=True)
    minheader = ctx.minimize_header(header)
    log.verbose("Bestrefs header:\n", log.PP(minheader))
    if condition:
        minheader = utils.condition_header(minheader)
    return ctx.get_best_references(minheader, include=include)


def test():
    """Run module doctests."""
    import doctest
    from crds.core import rmap
    return doctest.testmod(rmap)

if __name__ == "__main__":
    print(test())



"""This module contains functions related to generating and comparing the time order
of CRDS file names.   Different projects have different naming conventions.
"""
import os
import re

from . import config, log, utils
from .exceptions import NameComparisonError
from .constants import ALL_OBSERVATORIES

# =============================================================================================================

__all__ = [
    "newer",
    "generate_unique_name",
]

# =============================================================================================================

def generate_unique_name(filename, observatory=None, now=None):
    """Given and original filepath `filename` generate a unique CRDS name for the file."""
    if observatory is None:
        locator = utils.file_to_locator(filename)
    else:
        locator = utils.get_locator_module(observatory)
    return locator.generate_unique_name(filename, now)

# =============================================================================================================

def newer(name1, name2):
    """Determine if `name1` is a more recent file than `name2` accounting for
    limited differences in naming conventions. Official CDBS and CRDS names are
    comparable using a simple text comparison,  just not to each other.

    >>> newer("s7g1700gl_dead.fits", "hst_cos_deadtab_0001.fits")
    False

    >>> newer("hst_cos_deadtab_0001.fits", "s7g1700gl_dead.fits")
    True

    >>> newer("s7g1700gl_dead.fits", "bbbbb.fits")
    True

    >>> newer("bbbbb.fits", "s7g1700gl_dead.fits")
    False

    >>> newer("hst_cos_deadtab_0001.rmap", "hst_cos_deadtab_0002.rmap")
    False

    >>> newer("hst_cos_deadtab_0002.rmap", "hst_cos_deadtab_0001.rmap")
    True

    >>> newer("hst_cos_deadtab_0001.asdf", "hst_cos_deadtab_0050.fits")
    True

    >>> newer("hst_cos_deadtab_0051.fits", "hst_cos_deadtab_0050.asdf")
    False

    >>> newer("hst_cos_deadtab_0001.fits", "hst_cos_deadtab_99991.fits")
    False

    >>> newer("hst_cos_deadtab_99991.fits", "hst_cos_deadtab_0001.fits")
    True


    >>> newer("07g1700gl_dead.fits", "s7g1700gl_dead.fits")
    True

    >>> newer("s7g1700gl_dead.fits", "07g1700gl_dead.fits")
    False

    >>> newer("N/A", "s7g1700gl_dead.fits")
    False

    >>> newer("07g1700gl_dead.fits", "N/A")
    True

    >>> newer("N/A", "hst_cos_deadtab_0002.rmap")
    False
    >>> newer("hst_cos_deadtab_0002.rmap", "N/A")
    True

    >>> newer("16n1832tm_tmc.fits", "06n1832tm_tmc.fits")
    True
    >>> newer("06n1832tm_tmc.fits", "16n1832tm_tmc.fits")
    False
    >>> newer("06n1832tm_tmc.fits", "16n1832tm_tmg.fits")
    False

    >>> newer("hst_cos_deadtab_0001.fits", "16n1832tm_tmg.fits")
    Traceback (most recent call last):
    ...
    NameComparisonError: Unhandled name comparison case:  ('crds', 'synphot')

    >>> newer("07g1700gl_dead.fits", "16n1832tm_tmg.fits")
    Traceback (most recent call last):
    ...
    NameComparisonError: Unhandled name comparison case:  ('newcdbs', 'synphot')

    >>> newer("16n1832tm_tmg.fits", "7g1700gl_dead.fits")
    Traceback (most recent call last):
    ...
    NameComparisonError: Unhandled name comparison case:  ('synphot', 'oldcdbs')

    >>> newer("hst_cos_deadtab_0001.fits", "17g1700gl_dead.fits")
    Traceback (most recent call last):
    ...
    NameComparisonError: Unhandled name comparison case:  ('crds', 'newcdbs')

    >>> newer("17g1700gl_dead.fits", "hst_cos_deadtab_0001.fits")
    Traceback (most recent call last):
    ...
    NameComparisonError: Unhandled name comparison case:  ('newcdbs', 'crds')

    """
    cases = {
        ("crds", "crds")  : "compare_crds",
        ("oldcdbs", "oldcdbs") : "compare",
        ("newcdbs", "newcdbs") : "compare",
        ("oldsynphot", "oldsynphot") : "compare",
        ("newsynphot", "newsynphot") : "compare",

        ("crds", "oldcdbs") : True,
        ("oldcdbs", "crds") : False,

        ("newcdbs", "oldcdbs") : True,
        ("oldcdbs", "newcdbs") : False,

        ("newsynphot", "oldsynphot") : True,
        ("oldsynphot", "newsynphot") : False,
    }
    name1, name2 = crds_basename(name1), crds_basename(name2)
    class1 = classify_name(name1)
    class2 = classify_name(name2)
    case = cases.get((class1, class2), "raise")
    if name1 == "N/A":
        return False
    elif name2 =="N/A":
        result = True
    elif case == "compare_crds":
        if extension_rank(name1) == extension_rank(name2):
            serial1, serial2 = newstyle_serial(name1), newstyle_serial(name2)
            result = serial1 > serial2   # same extension compares by counter
        else:
            result = extension_rank(name1) > extension_rank(name2)
    elif case == "compare":
        result = name1 > name2
    elif case in [True, False]:
        result = case
    else:
        raise NameComparisonError("Unhandled name comparison case: ", repr((class1, class2)))
    log.verbose("Comparing filename time order:", repr(name1), ">", repr(name2), "-->", result)
    return result

def crds_basename(name):
    """basename() accounting for N/A pass thru."""
    if name == "N/A":
        return "N/A"
    else:
        return os.path.basename(name)

def classify_name(name):
    """Classify filename `name` as "crds", "synphot", "oldcdbs", or "newcdbs".

    >>> classify_name("jwst_miri_dark_0057.fits")
    'crds'

    >>> classify_name("s7g1700gl_dead.fits")
    'oldcdbs'

    >>> classify_name("07g1700gl_dead.fits")
    'newcdbs'

    >>> classify_name("16n1832tm_tmc.fits")
    'newsynphot'

    >>> classify_name("z6n1832tm_tmc.fits")
    'oldsynphot'

    Ad hoc names are generally mistakes which are immediately replaced.   Classify them
    as oldcdbs (oldest) so that comparison with anything else is interpreted as a
    replacement with something newer,  and nominal.

    >>> classify_name("bbbbbb.fits")
    'oldcdbs'
    """
    if name == "N/A":
        return "crds"
    elif crds_name(name):
        return "crds"
    elif old_synphot_name(name):
        return "oldsynphot"
    elif new_synphot_name(name):
        return "newsynphot"
    elif old_cdbs_name(name):
        return "oldcdbs"
    elif new_cdbs_name(name):
        return "newcdbs"
    else:
        # raise NameComparisonError("Failed to classify name", repr(name), "for determining time order.")
        log.warning("Failed to classify name", repr(name), "for determining time order.")
        return "oldcdbs"  # if it's a garbage name, make it oldest so replacing it is normal.

def crds_name(name):
    """Return True IFF `name` is a CRDS-style name, e.g. hst_acs.imap

    >>> crds_name("s7g1700gl_dead.fits")
    False
    >>> crds_name("1c82030ml_dead.fits")
    False
    >>> crds_name("hst.pmap")
    True
    >>> crds_name("hst_acs_darkfile_0001.fits")
    True
    >>> crds_name("16n1832tm_tmc.fits")
    False
    """
    return name.startswith(tuple(ALL_OBSERVATORIES))

OLD_CDBS = re.compile(config.complete_re(r"[A-Za-z][A-Za-z0-9]{8}_[A-Za-z0-9]{1,8}\.[A-Za-z0-9]{1,6}"))
NEW_CDBS = re.compile(config.complete_re(r"[0-9][A-Za-z0-9]{8}_[A-Za-z0-9]{1,8}\.[A-Za-z0-9]{1,6}"))

def old_cdbs_name(name1):
    """Return True IFF name1 is and original CDBS-style name.

    >>> old_cdbs_name("s7g1700gl_dead.fits")
    True
    >>> old_cdbs_name("1c82030ml_dead.fits")
    False
    >>> old_cdbs_name("16n1832tm_tmc.fits")
    False
    >>> old_cdbs_name("z6n1832tm_tmc.fits")
    True
    >>> old_cdbs_name("hst.pmap")
    False
    >>> old_cdbs_name("hst_acs_darkfile_0001.fits")
    False
    """
    return OLD_CDBS.match(name1) is not None

def new_cdbs_name(name1):
    """Return True IFF name1 is an extended CDBS-style name.

    >>> new_cdbs_name("s7g1700gl_dead.fits")
    False
    >>> new_cdbs_name("1c82030ml_dead.fits")
    True
    >>> new_cdbs_name("16n1832tm_tmc.fits")
    True
    >>> new_cdbs_name("z6n1832tm_tmc.fits")
    False
    >>> new_cdbs_name("hst.pmap")
    False
    >>> new_cdbs_name("hst_acs_darkfile_0001.fits")
    False
    """
    return NEW_CDBS.match(name1) is not None

NEW_SYNPHOT_RE = re.compile(config.complete_re(r"[0-9][a-zA-Za-z0-9]{7}m_(tmc|tmg|tmt)\.[A-Za-z0-9]{1,6}"))
OLD_SYNPHOT_RE = re.compile(config.complete_re(r"[A-Za-z][a-zA-Za-z0-9]{7}m_(tmc|tmg|tmt)\.[A-Za-z0-9]{1,6}"))

def synphot_name(name):
    """Return True IFF `name` is the name of an ETC master table file of some kind.

    >>> synphot_name("s7g1700gl_dead.fits")
    False
    >>> synphot_name("1c82030ml_dead.fits")
    False
    >>> synphot_name("16n1832tm_tmc.fits")
    True
    >>> synphot_name("16n1832tm_tmt.fits")
    True
    >>> synphot_name("16n1832tm_tmg.fits")
    True
    >>> synphot_name("hst.pmap")
    False
    >>> synphot_name("hst_acs_darkfile_0001.fits")
    False
    """
    return old_synphot_name(name) or new_synphot_name(name)

def new_synphot_name(name):
    """Return True IFF `name` is the name of an ETC master table file of some kind."""
    return NEW_SYNPHOT_RE.match(name) is not None

def old_synphot_name(name):
    """Return True IFF `name` is the name of an ETC master table file of some kind."""
    return OLD_SYNPHOT_RE.match(name) is not None

def extension_rank(filename):
    """Return a date ranking for `filename` based on extension, lowest numbers are oldest.

    >>> extension_rank("fooo.r0h")
    0.0
    >>> extension_rank("fooo.fits")
    1.0
    >>> extension_rank("fooo.json")
    2.0
    >>> extension_rank("fooo.yaml")
    3.0
    >>> extension_rank("/some/path/fooo.asdf")
    4.0
    """
    ext = os.path.splitext(filename)[-1]
    if re.match(r"\.r\dh", ext) or re.match(r"\.r\dd", ext):
        return 0.0
    elif ext == ".fits":
        return 1.0
    elif ext == ".json":
        return 2.0
    elif ext == ".yaml":
        return 3.0
    elif ext == ".asdf":
        return 4.0
    else:
        return 5.0

def newstyle_serial(name):
    """Return the serial number associated with a CRDS-style name.

    >>> newstyle_serial("hst_0042.pmap")
    42
    >>> newstyle_serial("hst_0990999.pmap")
    990999
    >>> newstyle_serial("hst_cos_0999.imap")
    999
    >>> newstyle_serial("hst_cos_darkfile_0998.fits")
    998
    >>> newstyle_serial("hst.pmap")
    -1
    """
    assert isinstance(name, str)
    serial_search = re.search(r"_(\d+)\.\w+", name)
    if serial_search:
        return int(serial_search.groups()[0], 10)
    else:
        return -1



"""Build-4 ACS hooks,  with modified for plug-in interface.

Apply to all ACS rmaps unless explicitly "un-hooked".

Dataset header preconditioning implemented as match-time operation.

Match fallbacks implemented as rmap-generation-time weaker matches,  single CRDS lookup.
"""
from .acs_common import *

def precondition_header_acs_biasfile_v1(rmap, header_in):
    """Mutate the incoming dataset header based upon hard coded rules
    and the header's contents.   This is an alternative to generating
    an equivalent and bulkier rmap.
    """
    header = dict(header_in)
    log.verbose("acs_biasfile_precondition_header:", log.format_parameter_list(header))
    exptime = timestamp.reformat_date(header["DATE-OBS"] + " " + header["TIME-OBS"])
    if (exptime < SM4):
        #if "APERTURE" not in header or header["APERTURE"] == "UNDEFINED":
        log.verbose("Mapping pre-SM4 APERTURE to N/A")
        header["APERTURE"] = "N/A"
    try:
        numcols = int(float(header["NUMCOLS"]))
    except ValueError:
        log.verbose("acs_biasfile_selection: bad NUMCOLS.")
        # sys.exc_clear()
    else:
        header["NUMCOLS"] = utils.condition_value(str(numcols))
        # if pre-SM4 and NUMCOLS > HALF_CHIP
        exptime = timestamp.reformat_date(header["DATE-OBS"] + " " + header["TIME-OBS"])
        if (exptime < SM4):
            if numcols > ACS_HALF_CHIP_COLS:
                if header["CCDAMP"] in ["A","D"]:
                    log.verbose("acs_bias_file_selection: exposure is pre-SM4, converting amp A or D " +
                                "to AD for NUMCOLS = " + header["NUMCOLS"])
                    header["CCDAMP"] = "AD"
                elif header["CCDAMP"] in ["B","C"]:
                    log.verbose("acs_bias_file_selection: exposure is pre-SM4, converting amp B or C " +
                                "to BC for NUMCOLS = " + header["NUMCOLS"])
                    header["CCDAMP"] = "BC"
    if header['DETECTOR'] == "WFC" and \
        header['XCORNER'] == "0.0" and header['YCORNER'] == "0.0":
        log.verbose("acs_biasfile_selection: precondition_header halving NUMROWS")
        try:
            numrows = int(float(header["NUMROWS"])) / 2
        except ValueError:
            log.verbose("acs_biasfile_selection: bad NUMROWS.")
            # sys.exc_clear()
        else:
            header["NUMROWS"] = utils.condition_value(str(numrows))
    return header     # XXXXXX RETURN NOW !!!!


# ===========================================================================

#   This section contains matching customizations.

# (('DETECTOR', 'CCDAMP', 'CCDGAIN', 'APERTURE', 'NUMCOLS', 'NUMROWS', 'LTV1', 'LTV2', 'XCORNER', 'YCORNER', 'CCDCHIP'), ('DATE-OBS', 'TIME-OBS')),

"""
def _fallback_biasfile(header_in):
    header = _precondition_header_biasfile(header_in)
    log.verbose("No matching BIAS file found for",
               "NUMCOLS=" + repr(header['NUMCOLS']),
               "NUMROWS=" + repr(header['NUMROWS']),
               "LTV1=" + repr(header['LTV1']),
               "LTV2=" + repr(header['LTV2']))
    log.verbose("Trying full-frame default search")
    if header['DETECTOR'] == "WFC":
        header["NUMCOLS"] = "4144.0"
        header["NUMROWS"] = "2068.0"
        header["LTV1"] = "24.0"
        header["LTV2"] = "0.0"
    else:
        header["NUMCOLS"] = "1062.0"
        header["NUMROWS"] = "1044.0"
        header["LTV1"] = "19.0"
        if header['CCDAMP'] in ["C","D"]:
            header["LTV2"] = "0.0"
        else: # assuming HRC with CCDAMP = A or B
            header["LTV2"] = "20.0"
    return header

def fallback_header(rmap, header):
    return None
    if rmap.filekind == "biasfile":
        # log.info("x", end="",sep="")
        return _fallback_biasfile(header)
    else:
        None
"""

#  Originally this was code used to update the ACS BIASFILE rmap with a fallback case.

def rmap_update_headers_acs_biasfile_v1(rmap, header_in):
    header = _precondition_header_biasfile(header_in)

    if header_matches(header, dict(DETECTOR='WFC', NUMCOLS='4144.0', NUMROWS='2068.0', LTV1='24.0', LTV2='0.0')):
        return dont_care(header, ['NUMROWS','NUMCOLS','LTV1', 'LTV2'])

    if header_matches(header, dict(DETECTOR='HRC', CCDAMP='C', NUMROWS='1044.0', NUMCOLS='1062.0', LTV1='19.0', LTV2='0.0')):
        return dont_care(header, ['NUMROWS','NUMCOLS','LTV1', 'LTV2'])

    if header_matches(header, dict(DETECTOR='HRC', CCDAMP='D', NUMROWS='1044.0', NUMCOLS='1062.0', LTV1='19.0', LTV2='0.0')):
        return dont_care(header, ['NUMROWS','NUMCOLS','LTV1', 'LTV2'])

    if header_matches(header, dict(DETECTOR='HRC', CCDAMP='C|D', NUMROWS='1044.0', NUMCOLS='1062.0', LTV1='19.0', LTV2='0.0')):
        return dont_care(header, ['NUMROWS','NUMCOLS','LTV1', 'LTV2'])

    if header_matches(header, dict(DETECTOR='HRC', CCDAMP='A', NUMROWS='1044.0', NUMCOLS='1062.0', LTV1='19.0', LTV2='20.0')):
        return dont_care(header, ['NUMROWS','NUMCOLS','LTV1', 'LTV2'])

    if header_matches(header, dict(DETECTOR='HRC', CCDAMP='B', NUMROWS='1044.0', NUMCOLS='1062.0', LTV1='19.0', LTV2='20.0')):
        return dont_care(header, ['NUMROWS','NUMCOLS','LTV1', 'LTV2'])

    if header_matches(header, dict(DETECTOR='HRC', CCDAMP='A|B', NUMROWS='1044.0', NUMCOLS='1062.0', LTV1='19.0', LTV2='20.0')):
        return dont_care(header, ['NUMROWS','NUMCOLS','LTV1', 'LTV2'])

    return header

def header_matches(header, conditions):
    for var, val in conditions.items():
        if header[var] != val:
            return False
    return True

def dont_care(header, vars):
    header = dict(header)
    for var in vars:
        header[var] = "N/A"
    return header

# =============================================================================================

# This section contains rmap generation code.

header_additions = [   # dictionary items (ordered)
]

def acs_biasfile_filter(kmap):
    """APERTURE was added late as a matching parameter and so many existing references
    have an APERTURE value of '' in CDBS.   Where it's relevant,  it's actually defined.
    Here we change '' to * to make CRDS ignore it when it doesn't matter.   We also change
    APERTURE to * for any useafter date which precedes SM4 (possibly they define APERTURE).

    add_fallback_to_kmap() duplicates the correct filemaps to simulate the fallback header lookup.
    """
    replacement = "*"
    log.info("Hacking ACS biasfile  APERTURE macros.  Changing APERTURE='' to APERTURE='%s'" % replacement)
    start_files = total_files(kmap)
    for match, fmaps in kmap.items():
        new_key = na_key(match, replacement)
        if match[3] == '':
            if new_key not in kmap:
                kmap[new_key] = []
            kmap[new_key] = sorted(set(kmap[new_key] + fmaps))
            del kmap[match]
            for fmap in fmaps:
                log.info("Unconditionally mapping APERTURE '' to '%s' for" % replacement, fmap)
            continue
        remap_fmaps = []
        for fmap in fmaps[:]:
            if fmap.date < SM4:
                log.info("Remapping <SM4 APERTURE to '%s'" % replacement, repr(fmap))
                remap_fmaps.append(fmap)
                fmaps.remove(fmap)
        if remap_fmaps:
            if new_key not in kmap:
                kmap[new_key] = []
            kmap[new_key].extend(remap_fmaps)
            log.info("Moving", match, "to", new_key, "for files", total_files({None:remap_fmaps}))
            log.info("Remainder", match, "=", total_files({None:kmap[match]}))
            # log.info("New kmap for", repr(new_key), "is", repr(kmap[new_key]))
        if not fmaps and (new_key != match):
            del kmap[match]

    dropped_files = start_files - total_files(kmap)
    if dropped_files:  # bummer,  bug in my code...
        log.error("Dropped files:", sorted(dropped_files))

    kmap = add_fallback_to_kmap(kmap,
        matches=dict(DETECTOR='WFC', NUMCOLS='4144.0', NUMROWS='2068.0', LTV1='24.0', LTV2='0.0'),
        dont_care=['NUMROWS','NUMCOLS','LTV1', 'LTV2'])

    kmap = add_fallback_to_kmap(kmap,
        matches=dict(DETECTOR='HRC', CCDAMP='C', NUMROWS='1044.0', NUMCOLS='1062.0', LTV1='19.0', LTV2='0.0'),
        dont_care=['NUMROWS','NUMCOLS','LTV1', 'LTV2'])

    kmap = add_fallback_to_kmap(kmap,
        matches=dict(DETECTOR='HRC', CCDAMP='D', NUMROWS='1044.0', NUMCOLS='1062.0', LTV1='19.0', LTV2='0.0'),
        dont_care=['NUMROWS','NUMCOLS','LTV1', 'LTV2'])

    kmap = add_fallback_to_kmap(kmap,
        matches=dict(DETECTOR='HRC', CCDAMP='A', NUMROWS='1044.0', NUMCOLS='1062.0', LTV1='19.0', LTV2='20.0'),
        dont_care=['NUMROWS','NUMCOLS','LTV1', 'LTV2'])

    kmap = add_fallback_to_kmap(kmap,
        matches=dict(DETECTOR='HRC', CCDAMP='B', NUMROWS='1044.0', NUMCOLS='1062.0', LTV1='19.0', LTV2='20.0'),
        dont_care=['NUMROWS','NUMCOLS','LTV1', 'LTV2'])

    return kmap, header_additions

#  An old example of hacking the kmap....
#     kmap[('UVIS', 'G280_AMPS', 1.5, 1.0, 1.0, 'G280-REF', 'T')] = \
#       [rmap.Filemap(date='1990-01-01 00:00:00', file='t6i1733ei_bia.fits',
#               comment='Placeholder file. All values set to zero.--------------------------, 12047, Jun 18 2009 05:36PM')]
#    return kmap, header_additions

def na_key(match, replacement='*'):
    """Replace APERTURE with N/A or *"""
    new = list(match)
    new[3] = replacement
    return tuple(new)

def total_files(kmap):
    total = set()
    for match, fmaps in kmap.items():
        total = total.union({fmap.file for fmap in fmaps})
    return total

def add_fallback_to_kmap(kmap, matches, dont_care,
    parkeys=('DETECTOR', 'CCDAMP', 'CCDGAIN', 'APERTURE', 'NUMCOLS', 'NUMROWS',
             'LTV1', 'LTV2', 'XCORNER', 'YCORNER', 'CCDCHIP')):
    """Copy items in `kmap` whose keys match the parameters in `matches`,  setting
    the key-copy values named in `dont_care` to 'N/A'.   The copy with some 'N/A's is a fallback.
    `parkeys` names the items of each tuple key in `kmap`,  in order.
    """
    kmap = defaultdict(list, kmap.items())
    for key in kmap.keys():
        if key_matches(key, parkeys, matches):
            new_key = set_dont_care(key, parkeys, dont_care)
#            if new_key not in kmap:
#                kmap[new_key] = []
            kmap[new_key].extend(list(kmap[key]))
            kmap[new_key].sort()
            log.info("Creating fallback", repr(key), "-->", repr(new_key))
    return kmap

def key_matches(key, parkeys, matches):
    """Return True IFF `key` matches all the values in dict `matches`.
    Corresponding values in tuple `key` are named by values in `parkeys`.
    """
    for i, name in enumerate(parkeys):
        if name in matches:
            if utils.condition_value(matches[name]) != utils.condition_value(key[i]):
                log.verbose("Exiting on", repr(name),
                         utils.condition_value(matches[name]),
                         utils.condition_value(key[i]))
                return False
    log.verbose("Matching", repr(key))
    return True

def set_dont_care(key, parkeys, dont_care):
    """Set the values of `key` named in `dont_care` to 'N/A'."""
    key = list(key)
    for i, name in enumerate(parkeys):
        if name in dont_care:
            key[i] = 'N/A'
    return tuple(key)

# =============================================================================================



"""This module provides a project independent interface for determining the sequence of
pipelines used to process the given dataset filepath or dataset with a particular header.

>>> header_to_pipelines(test_header("0.7.0", "FGS_DARK"))
['calwebb_dark.cfg', 'skip_2b.cfg']

>>> header_to_pipelines(test_header("0.7.0", "NRS_BRIGHTOBJ"))
['calwebb_sloper.cfg', 'calwebb_spec2.cfg']

>>> header_to_pipelines(test_header("0.7.0", "MIR_IMAGE"))
['calwebb_sloper.cfg', 'calwebb_image2.cfg']

>>> header_to_pipelines(test_header("0.7.0", "MIR_LRS-FIXEDSLIT"))
['calwebb_sloper.cfg', 'calwebb_spec2.cfg']

"""
from crds.core import utils
from crds import data_file

# --------------------------------------------------------------------------------------

def test_header(calver, exp_type):
    """Create a header-like dict from `calver` and `exp_type` to support
    testing.
    """
    header = {
        "META.INSTRUMENT.NAME" : "SYSTEM",
        "REFTYPE" : "CRDSCFG",
        "META.CALIBRATION_SOFTWARE_VERSION" : calver,
        "META.EXPOSURE.TYPE" : exp_type,
        }
    return header

# --------------------------------------------------------------------------------------

def filepath_to_pipelines(dataset_path, context=None):
    """Given `dataset_path` and `context` name,  return the sequence of calibration
    code pipeline .cfg files nominally required to process `dataset_path`.
    """
    header = data_file.get_header(dataset_path, context)
    return header_to_pipelines(header, context)

def header_to_pipelines(header, context=None):
    """Based on `header` and `context` provide the project specific list of calibration
    code pipelines (.cfg files) used to process a dataset with `header`.
    """
    locator = utils.header_to_locator(header)
    return locator.header_to_pipelines(header, context)



"""This module consolidates a number of aspects of reference type definition which were
originally reverse engineered from CDBS and several different spec files.  It is
organized around loading type specs or prototype rmaps from the "specs" subdirectory of
an observatory/subsystem package.   For HST this reduces defining new types to adding
a prototype rmap and defining .tpn files in the observatory package.
"""
import os.path
import collections
import glob
import json

from . import rmap, log, utils, exceptions
from . import generic_tpn

# =============================================================================
#  Global table loads used at operational runtime:

def _invert_instr_dict(dct):
    """Invert a set of nested dictionaries of the form {instr: {key: val}}
    to create a dict of the form {instr: {val: key}}.
    """
    inverted = {}
    for instr in dct:
        inverted[instr] = utils.invert_dict(dct[instr])
    return inverted

# =============================================================================
'''
def write_spec(specpath, spec_dict):
    """Pretty print a spec dictionary."""
    with open(specpath, "w+") as spec:
        spec.write("{\n")
        for key, value in sorted(spec_dict.items()):
            spec.write("    {} : {},\n".format(repr(key), repr(value)))
        spec.write("}\n")
'''

# =============================================================================

class TypeSpec(dict):
    """This class captures type definition parameters for a single type"""

    def __init__(self, header):
        """Initialize this TypeSpec from dict `header`,  enforcing requirments and creating suitable
        defaults for missing fields.
        """
        header = utils.Struct(header)
        if "suffix" not in header:
            raise exceptions.CrdsTypeSpecError("Missing 'suffix' field in type spec.")
        if "filetype" not in header:
            raise exceptions.CrdsTypeSpecError("Missing 'filetype' field in type spec.")
        if  "text_descr" not in header and  header.filetype != "all":
            raise exceptions.CrdsTypeSpecError("Missing 'text_descr' field in type spec.")
        if "tpn" not in header:
            header.tpn = header.instrument.lower() + "_" + header.suffix + ".tpn"
        if "ld_tpn" not in header:
            header.ld_tpn = header.instrument.lower() + "_" + header.suffix + "_ld.tpn"
        if "file_ext" not in header:
            header.file_ext = ".fits"
        if "unique_rowkeys" not in header:
            header.unique_rowkeys = None
        if "extra_keys" not in header:
            header.extra_keys = None
        super(TypeSpec, self).__init__(header.items())

    @classmethod
    def from_file(cls, filename):
        """For historical HST types,  build type info from a spec file derived from CDBS specs like
        reference_file_defs.xml or cdbscatalog.dat.  For new CRDS-only types,  use a prototype rmap
        with an enhanced header to define the type.   Prototypes should be submissible but should not
        contain references.
        """
        log.verbose("Loading type spec", repr(filename), verbosity=75)
        if filename.endswith(".spec"):
            return cls(utils.evalfile(filename))
        else:
            return cls(rmap.load_mapping(filename).header)

# =============================================================================

def load_specs(spec_path):
    """Load either the combined .json formatted type specs (preferred) or specs from
    individual type spec files (if necessary).
    """
    combined_specs_path = os.path.join(spec_path, "combined_specs.json")
    if os.path.exists(combined_specs_path):
        headers = load_json_specs(combined_specs_path)
    else:
        headers = load_raw_specs(spec_path)
        with log.warn_on_exception("Failed to save type specs .json at", repr(combined_specs_path)):
            save_json_specs(headers, combined_specs_path)
    type_specs = { instr :
                   { filekind : TypeSpec(headers[instr][filekind])
                     for filekind in headers[instr]
                     }
                   for instr in headers
                   }
    return type_specs

def load_raw_specs(spec_path):
    """Return a dictionary of TypeSpecs loaded from directory `spec_path` of form:

    { instrument : { filetype : header_dict, ...}, ... }
    """
    with log.error_on_exception("Failed loading type specs from:", repr(spec_path)):
        specs = collections.defaultdict(dict)
        for spec in glob.glob(os.path.join(spec_path, "*.spec")) + glob.glob(os.path.join(spec_path, "*.rmap")):
            instr, reftype = os.path.splitext(os.path.basename(spec))[0].split("_")
            with log.error_on_exception("Failed loading", repr(spec)):
                specs[instr][reftype] = dict(TypeSpec.from_file(spec))
        return specs
    return {}

def save_json_specs(specs, combined_specs_path):
    """Write out the specs dictionary returned by _load_specs() as .json in one combined file."""
    specs_json = json.dumps(specs, indent=4, sort_keys=True, separators=(',', ':'))
    with open(combined_specs_path, "w+") as specs_file:
        specs_file.write(specs_json)
        log.info("Saved combined type specs to", repr(combined_specs_path))

def load_json_specs(combined_specs_path):
    """Load specs from the .json combined spec file under `specs_path`."""
    with open(combined_specs_path, "r") as combined_specs_file:
        return json.load(combined_specs_file)

# =============================================================================

def from_package_file(observatory, pkg):
    """Given the __file__ from a package,  load specs from the package's specs subdirectory
    and construct and return a TypeParameters object.
    """
    here = (os.path.dirname(pkg) or ".")
    specs_path = os.path.join(here, "specs")
    unified_defs = load_specs(specs_path)
    return TypeParameters(observatory, unified_defs)

class TypeParameters:
    """Inialized from a dictionary of TypeSpec's from load_specs(), compute observatory enumerations
    and type field inter-relationships and cache them as attributes.
    """
    def __init__(self, observatory, unified_defs):

        self.observatory = observatory
        self.unified_defs = unified_defs

        sorted_udef_items = sorted(unified_defs.items())

        with log.error_on_exception("Can't determine instruments from specs."):
            self.instruments = [instr.lower() for instr in sorted(self.unified_defs.keys())]

        with log.error_on_exception("Can't determine types from specs."):
            self.filekinds = sorted(
                set(reftype.lower() for instr, reftypes in sorted_udef_items
                    for reftype in reftypes))

        with log.error_on_exception("Can't determine extensions from specs."):
            self.extensions = sorted(
                set(params.get("file_ext", ".fits") for instr, reftypes in sorted_udef_items
                    for reftype, params in reftypes.items())) + [".pmap", ".imap", ".rmap"]

        with log.error_on_exception("Can't determine type text descriptions from specs."):
            self.text_descr = {
                reftype.lower() : params["text_descr"]
                for instr, reftypes in sorted_udef_items
                for reftype, params in reftypes.items()
                }

        with log.error_on_exception("Failed determining filekind_to_suffix"):
            self._filekind_to_suffix = {
                instr : {
                    filekind.lower() : self.unified_defs[instr][filekind]["suffix"].lower()
                    for filekind in self.unified_defs[instr]
                    }
                for instr in self.unified_defs
                }

        with log.error_on_exception("Failed determining suffix_to_filekind"):
            self._suffix_to_filekind = _invert_instr_dict(self._filekind_to_suffix)

        with log.error_on_exception("Failed determining filetype_to_suffix"):
            self._filetype_to_suffix = {
                instr : {
                    self.unified_defs[instr][filekind]["filetype"].lower() : self.unified_defs[instr][filekind]["suffix"].lower()
                    for filekind in self.unified_defs[instr]
                    }
                for instr in self.unified_defs
                }

        with log.error_on_exception("Failed determining suffix_to_filetype"):
            self.suffix_to_filetype = _invert_instr_dict(self._filetype_to_suffix)

        with log.error_on_exception("Failed determining unique_rowkeys"):
            self.row_keys = {
                instr : {
                    filekind.lower() : self.unified_defs[instr][filekind]["unique_rowkeys"]
                    for filekind in self.unified_defs[instr]
                    }
                for instr in self.unified_defs
                }

    @property
    def locator(self):
        """Deferred evaluation do to import orders."""
        return utils.get_locator_module(self.observatory)  # pluggable by observatory

    def filetype_to_filekind(self, instrument, filetype):
        """Map the value of a FILETYPE keyword onto it's associated
        keyword name,  i.e.  'dark image' --> 'darkfile'
        """
        instrument = instrument.lower()
        filetype = filetype.lower()
        if instrument == "nic":
            instrument = "nicmos"
        suffix = self._filetype_to_suffix[instrument][filetype]
        return self._suffix_to_filekind[instrument][suffix]

    def suffix_to_filekind(self, instrument, suffix):
        """Map the value of an instrument and TPN suffix onto it's
        associated filekind keyword name,  i.e. drk --> darkfile
        """
        if instrument == "nic":
            instrument = "nicmos"
        return self._suffix_to_filekind[instrument][suffix]

    def filekind_to_suffix(self, instrument, filekind):
        """Given and instrument (e.g. acs) and filekind (e.g. darkfile)
        turn the associated file suffix (e.g. drk) for use in generating
        legacy filenames.
        """
        if instrument == "nic":
            instrument = "nicmos"
        return self._filekind_to_suffix[instrument][filekind]

# =============================================================================

    @utils.cached
    def get_all_tpn_paths(self, instrument, filekind, field):
        """Return all the TpnInfo objects defined by the specified parameters.
        Doesn't include "extra" TpnInfo's defined by a project for a specific
        reference file.
        """
        return [ self.locator.tpn_path(tpn_file)
                 for tpn_file in self.reference_props_to_validator_keys(
                         instrument, filekind, field=field) ]

    @utils.cached
    def get_all_tpninfos(self, instrument, filekind, field):
        """Return all the TpnInfo objects defined by the specified parameters.
        Doesn't include "extra" TpnInfo's defined by a project for a specific
        reference file.
        """
        tpns = []
        for tpn_path in self.get_all_tpn_paths(instrument, filekind, field):
            tpns.extend(generic_tpn.get_tpninfos(tpn_path))
        return sorted(list(set(tpns)))


    def reference_props_to_validator_keys(self, instrument, filekind, field="tpn"):
        """Return the sequence of validator keys associated with `filename`.   A validator key
        is nominally a .tpn filename and can vary by observatory, instrument, and type as well
        as by functions on the header of `filename`.
        """
        results = []
        def append_tpn_level(results, instrument, filekind):
            """Append the validator key for one level of the `instrument`
            and `filekind` mutating list `results`.
            """
            try:
                tpnfile = self.unified_defs[instrument][filekind][field]
                log.verbose("Adding validator key", repr(tpnfile), verbosity=70)
                results.append(tpnfile)
            except Exception as exc:
                log.verbose_warning("Can't find TPN key for",
                    (instrument, filekind, field), ":", str(exc),
                                    verbosity=75)
        append_tpn_level(results, instrument, filekind)
        append_tpn_level(results, instrument, "all")
        append_tpn_level(results, "all", filekind)
        append_tpn_level(results, "all", "all")
        return results

# -----------------------------------------------------------------------------

    def reference_name_to_tpninfos(self, filename, field="tpn"):
        """Return the .tpn file key associated with reference `filename`.
        Strictly speaking this should be driven by mapping_validator_key...  but the interface
        for that is wrong so slave it to reference_name_to_tpn_key instead,  historically
        one-for-one.
        """
        filepath = self.locator.locate_file(filename)
        instrument, filekind = self.locator.get_file_properties(filepath)
        return self.get_all_tpninfos(instrument, filekind, field=field)

# -----------------------------------------------------------------------------

    def reference_name_to_tpn_text(self, filename, field="tpn"):
        """Given reference `filename`,  return the text of the corresponding .tpn"""
        instrument, filekind = self.locator.get_file_properties(filename)
        lines = []
        for tpn in reversed(self.get_all_tpn_paths(instrument, filekind, field)):
            with log.verbose_warning_on_exception(
                    "Failed loading tpn", repr(tpn)):
                label = "From TPN: " + os.path.basename(tpn)
                underline = len(label) * "-"
                lines += [ label, underline ]
                lines += generic_tpn.load_tpn_lines(tpn)
                lines += [""]
        # Extra tpns from datamodels have no .tpn serialization
        # hence displayed a TpnInfo object reprs.
        extra_tpns = self.locator.get_extra_tpninfos(filename)
        if  extra_tpns:
            extras_title = f"From {self.observatory} datamodels:"
            underline = "-"*len(extras_title)
            lines += [extras_title, underline]
            for tpn in extra_tpns:
                lines += [repr(tpn)]
        return "\n".join(lines) + "\n"

    def reference_name_to_ld_tpn_text(self, filename):
        """Given reference `filename`,  return the text of the corresponding _ld.tpn"""
        return self.reference_name_to_tpn_text(filename, "ld_tpn")

# =============================================================================

    def get_row_keys(self, instrument, filekind):
        """Return the row_keys which define unique table rows corresponding to mapping.

        These are used for "mode" checks to issue warnings when unique rows are deleted
        in a certify comparison check against the preceding version of a table.

        row_keys are now also utlized to perform "affected datasets" table row
        lookups which essentially requires emulating that aspect of the calibration
        software.  Consequently, row_keys now have a requirement for a higher level
        of fidelity since they were originally defined for mode checks, since the
        consequences of inadequate row keys becomes failed "affects checks" and not
        merely an extraneous warning.  In their capacity as affected datasets
        parameters,  row_keys must be supported by the interface which connects the
        CRDS server to the appropriate system dataset parameter database,  DADSOPS
        for HST.   That interface must be updated when row_keys.dat is changed.

        The certify mode checks have a shaky foundation since the concept of mode
        doesn't apply to all tables and sometimes "data" parameters are required to
        render rows unique.   The checks only issue warnings however so they can be
        ignored by file submitters.

        For HST calibration references mapping is an rmap.
        """
        try:
            return self.row_keys[instrument][filekind]
        except KeyError:
            log.verbose("No unique row keys defined for", repr((instrument, filekind)))
            return []

    def get_row_keys_by_instrument(self, instrument):
        """To support defining the CRDS server interface to DADSOPS, return the
        sorted list of row keys necessary to perform all the table lookups
        of an instrument.   These (FITS) keywords will be used to instrospect
        DADSOPS and identify table fields which provide the necessary parameters.
        """
        keyset = set()
        instrument = instrument.lower()
        for filekind in self.row_keys[instrument]:
            keyset |= set(self.row_keys[instrument][filekind] or [])
        return sorted([key.lower() for key in keyset])

    # =============================================================================

    def get_filekinds(self, instrument):
        """Return the sequence of filekind strings for `instrument`."""
        instrument = instrument.lower()
        return sorted(self._filekind_to_suffix[instrument].keys())

    def get_item(self, instrument, filekind, name):
        """Return config item `name` for `instrument` and `filekind`"""
        instrument = instrument.lower()
        filekind = filekind.lower()
        return self.unified_defs[instrument][filekind][name]

# =============================================================================

def get_types_object(observatory):
    """Each observatory package instantiates a types object from the observatory's
    type specs using generic code defined in this module;  any generic code related
    to types is defined here,  but ultimately the instantiate TYPES object is bound
    to a specific observatory and defined in that package.
    """
    pkg = utils.get_observatory_package(observatory)
    return getattr(pkg, "TYPES")



"""Master ACS hooks module,  importer of versioned hooks."""

from .acs_v1 import precondition_header_acs_biasfile_v1  # , acs_biasfile_filter
from .acs_v2 import precondition_header_acs_biasfile_v2, fallback_header_acs_biasfile_v2, acs_biasfile_filter, acs_darkfile_filter

#
# This section contains relevant code from cdbsquery.py and explanation,  such as it is.
# cdbsquery.py was/is part of CDBS.  This module is about capturing those quirks in different
# ways for CRDS.
#
'''
This is special case code which CDBS implements in cdbsquery.py
(see crds/hst/cdbs/cdbs_bestrefs/cdbsquery.py or get the latest from Mike Swam / OPUS)

Fundamentally, cdbsquery searches the reference file database
(reffile_ops...) which enumerates how specific dataset header
configurations map onto reference files.  Each row of reffile_ops maps
one instrument configuration and date onto the reference file which
handles it.  Some files are used in many rows.  Each column of a row
specifies the value of a particular parameter.

reffile_ops matching parameters, and other things,  are specified in the CDBS file
reference_file_defs.xml.   Parameters listed here are important in some way for
matching.

Ideally,  CDBS and CRDS implement something like:

   match(dataset parameters, reference parameters) --> reference.

Here dataset and reference correspond to the tuples of relevant header values
or database values.  1:1,  match them up,  shebang.

CDBS does the matching as a single SQL query against refile_ops...  cdbsquery
builds the SQL which constrains each parameter to the configuration of one dataset.

CRDS encodes the matching parameters of reference_file_ops in a
more compact way,  the rmaps.   The rmaps are interpreted by CRDS code
in selectors.py rather than by a SQL engine.

In practice,  CDBS special case code like that below does this:

   match(f(dataset parameters), reference parameters) -->  reference or g(dataset)

There are a few kinds of hacks:

1. The special case code excludes some dataset parameters from
matching in the reference database.  There might be no column for that
parameter but it is needed later.

2. The special case code fudges some of the dataset parameters,
replacing dataset[X] with f(dataset),  where the fudged value of one
matching parameter may depend on the value of many other parameters.

3. Sometimes fallback code,  g(dataset),  is used as a secondary solution
when the primary solution fails to obtain a match.

In CDBS:

"dataset" can come from the dadsops database (there are many servers and variants).

"dataset" can come from a FITS header.

At match time,  "reference" comes from the reffile_ops database.   Matching is
done by doing a gigantic anded SQL query which equates reffile_ops columns with
specific dataset values,  doing a gigantic anded SQL query,  and also choosing
the maximum applicable USEAFTER.  In cdbsquery,  the query is built one
parameter at a time.   Some parameters are omitted from the query.

At new reference submission time, "reference" comes from the ref file
header and is exploded in to refile_ops database rows.

In CRDS:

"dataset" comes from the same places, either a database or dataset file header.

At rmap generation time,  "reference" comes from the reffile_ops database and
is encoded into rmaps.

At new file submission time, "reference" comes from the reference file
headers and must be appropriately matched against rmaps to see where
to insert or replace files.   This is batch reference submission / refactoring.

At match time, essentially "reference" comes from the rmaps and
matching is done by recursively executing CRDS selector.py Selector code
on nested selectors.

# ================================================================================

  def find_exposure_start(self, aSource):
    # fill the exposure start time, after converting keyword value
    # to datetime format from source format
    beyond_SM4 = True    # default
    try:
      if aSource._keywords.has_key('TEXPSTRT'):
       #timetuple = time.strptime(aSource._keywords['TEXPSTRT'][0],
       #                        "%Y.%m.%d %H:%M:%S")
        timetuple = time.strptime(aSource._keywords['TEXPSTRT'][0].rstrip(),
                                "%Y.%m.%d %H:%M:%S")
      elif aSource._keywords.has_key('EXPSTART'):
       #timetuple = time.strptime(aSource._keywords['EXPSTART'][0],
       #                        "%Y.%m.%d %H:%M:%S ")
        timetuple = time.strptime(aSource._keywords['EXPSTART'][0].rstrip(),
                                "%Y.%m.%d %H:%M:%S")
      #
      # these are used in FITS input mode
      elif (aSource._keywords.has_key('PSTRTIME') and
            aSource._keywords['PSTRTIME'][0] != ' '):
        opusutil.PrintMsg("D","Trying PSTRTIME for exposure start:"+
                          aSource._keywords['PSTRTIME'][0]+'|')
        # its there, use it and convert it
       #temptuple = time.strptime(aSource._keywords['PSTRTIME'][0],
       #                        "%Y.%j:%H:%M:%S")
        temptuple = time.strptime(aSource._keywords['PSTRTIME'][0].rstrip(),
                                "%Y.%j:%H:%M:%S")
        # convert DOY to month and day
        parts = string.split(aSource._keywords['PSTRTIME'][0],":")
        moparts = string.split(parts[0],".")
        if int(moparts[0]) % 4 == 0:
          modays = (31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
        else:
          modays = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
        monum = 0
        doy = int(moparts[1])
        while doy > modays[monum]:
          doy = doy - modays[monum]
          monum = monum + 1
        # adjust month to 1-relative value
        month = monum + 1
        day = doy
        timetuple = (temptuple[0],month,day,temptuple[3],temptuple[4],
                     temptuple[5],temptuple[6],temptuple[7],temptuple[8])
      elif (aSource._keywords.has_key('TVSTART') and
            aSource._keywords['TVSTART'][0] != ' '):
        # thermal vac keyword for COS/WFC3 testing
        opusutil.PrintMsg("D","Trying TVSTART for exposure start:"+
                          aSource._keywords['TVSTART'][0]+'|')
        timetuple = time.strptime(aSource._keywords['TVSTART'][0],
                                  "%Y-%m-%dT%H:%M:%S")
      elif (aSource._keywords.has_key('DATE-OBS') and
            aSource._keywords['DATE-OBS'][0] != ' ' and
            aSource._keywords.has_key('TIME-OBS') and
            aSource._keywords['TIME-OBS'][0] != ' ') :
        opusutil.PrintMsg("D","Trying DATE-OBS,TIME-OBS for exposure start.")
        timetuple = time.strptime(aSource._keywords['DATE-OBS'][0]+' '+
                                  aSource._keywords['TIME-OBS'][0],
                                  "%Y-%m-%d %H:%M:%S")
      else:
        opusutil.PrintMsg("E","Failed to find any exposure start keys")
        raise FailedTimeConversion
      #
      exposure_start = time.strftime("%m/%d/%Y %H:%M:%S",timetuple)
      #
      # determine if exposure start is post SM4
      exposure_start_compare = time.strftime("%Y.%j:%H:%M:%S",timetuple)
      if (exposure_start_compare > SM4_YYYYDDDHHMMSS): beyond_SM4 = True
      else:                                            beyond_SM4 = False

    except Exception:
      traceback.print_exc()
      opusutil.PrintMsg("E","Failed time conversion for exposure start.")
      raise FailedTimeConversion
    #
    return exposure_start, beyond_SM4


  def acs_bias_file_selection(self, querynum, thereffile, aSource, beyond_SM4):
    querytxt = ""
    #
    # add the file selection fields
    for k in thereffile._file_selections:
      if k._restrictions:
        # see if the restriction allows this file selection field
        opusutil.PrintMsg("D",'found a file select restricted: '+
                               k._restrictions)
        if (not eval(k._restrictions)):    # CDBS code comment
          opusutil.PrintMsg("D","File_selection on "+k._field+
                                " restricted away")
          continue
      #
      # skip these (not used in selection, only for special tests)
      if (k._field == 'XCORNER' or k._field == 'YCORNER' or
          k._field == 'CCDCHIP'):
          continue
      #
      # skip APERTURE prior to SM4, otherwise use it (PR 72156)
      if (k._field == 'APERTURE' and not beyond_SM4):
          continue
      #
      # special cases
      elif k._field == 'NUMROWS':
        querytxt = querytxt + "and acs_row_"+querynum+ ".naxis2 = OBS_NAXIS2 "
        continue
      elif k._field == 'NUMCOLS':
        querytxt = querytxt + "and acs_row_"+querynum+ ".naxis1 = OBS_NAXIS1 "
        continue
      elif k._field == 'LTV1':
        querytxt = querytxt + "and acs_row_"+querynum+ ".ltv1 = OBS_LTV1 "
        continue
      elif k._field == 'LTV2':
        querytxt = querytxt + "and acs_row_"+querynum+ ".ltv2 = OBS_LTV2 "
        continue
      # only apply this case for exposures before SM4
      elif k._field == 'CCDAMP' and not beyond_SM4:
        #
        # convert amp A or D reads to AD if size is more than half a chip
        if ((aSource._keywords["CCDAMP"][0] == "A" or
             aSource._keywords["CCDAMP"][0] == "D") and
            aSource._keywords["NUMCOLS"][0] > ACS_HALF_CHIP_COLS):
            opusutil.PrintMsg("I","acs_bias_file_selection: exposure is pre-SM4, converting amp A or D "+
                                  "to AD for NUMCOLS = "+
                                  str(aSource._keywords["NUMCOLS"][0]))
            aSource._keywords["CCDAMP"][0] = "AD"
        #
        # convert amp B reads to BC if size is more than half a chip
        elif ((aSource._keywords["CCDAMP"][0] == "B" or
               aSource._keywords["CCDAMP"][0] == "C") and
              aSource._keywords["NUMCOLS"][0] > ACS_HALF_CHIP_COLS):
            opusutil.PrintMsg("I","acs_bias_file_selection: exposure is pre-SM4, converting amp B or C"+
                                  "to BC for NUMCOLS = "+
                                  str(aSource._keywords["NUMCOLS"][0]))
            aSource._keywords["CCDAMP"][0] = "BC"
      #
      # apply the file selection field
      # (first as a string, but if that fails, as a number converted to string,
      #  since numbers are not quoted but string are)
      try:
        querytxt = (querytxt + "and acs_row_"+querynum+"." +
                 string.lower(k._field) + " = '" +
                 aSource._keywords[k._field][0] + "' ")
      except TypeError:
        querytxt = (querytxt + "and acs_row_"+querynum+"." +
                 string.lower(k._field) + " = " +
                 str(aSource._keywords[k._field][0]) + " ")
    #
    return querytxt
'''



import os
import os.path
import sys
import importlib

import warnings

__rationale__ = "The __rationale__ variable is no longer maintained and will be removed in a future release."

warnings.filterwarnings(
    "ignore", ".*found in sys.modules after import of package.*")

# ============================================================================

__all__ = [
    "getrecommendations",
    "getreferences",
    "assign_bestrefs",
    "get_pickled_mapping",
    "get_symbolic_mapping",
    "get_cached_mapping",
    "get_context_name",
    "asmapping",
    "locate_mapping",
    "locate_file",
    "get_default_context",
    ]

# ============================================================================

from .core import config   # module

from .core.rmap import get_cached_mapping, asmapping
from .core.config import locate_mapping, locate_file

from .core import exceptions
from .core.exceptions import *
from .core.constants import ALL_OBSERVATORIES, INSTRUMENT_KEYWORDS

from .core.heavy_client import getreferences, getrecommendations
from .core.heavy_client import get_symbolic_mapping, get_pickled_mapping
from .core.heavy_client import get_context_name

from crds.client import api
from crds.client import get_default_context

from crds.bestrefs import assign_bestrefs

from ._version import version as __version__

# ============================================================================

'''This code section supports moving modules from the root crds namespace into
sub-packages while still supporting the original CRDS package external interface.
This allows the code to be partitioned without changing external imports.  This
is made more difficult by the inapplicability of namespace packages because this
__init__ is not empty.

The strategy employed here is to implement core packages normally in crds.core,
then alias them into the top level crds namespace using importlib.import_module()
and sys.modules to make it appear as if each core package has already been imported
and belongs to the top level namespace.
'''
def _alias_subpackage_module(subpkg, modules):
    """Alias each module from `modules` of `subpkg` to appear in this
    namespace.
    """
    for module in modules:
        globals()[module] = importlib.import_module(subpkg + "." + module)
        sys.modules["crds." + module] = sys.modules[subpkg + "." + module]

_CORE_MODULES = [
    "pysh",
    "python23",
    "exceptions",
    "log",
    "config",
    "constants",
    "utils",
    "crds_cache_locking",
    "timestamp",
    "custom_dict",
    "selectors",
    "mapping_verifier",
    "substitutions",
    "rmap",
    "heavy_client",
    "cmdline",
    "naming",
    "git_version",
]

# e.g. make crds.rmap importable same as crds.core.rmap reorganized code
_alias_subpackage_module("crds.core", _CORE_MODULES)

# ============================================================================

# e.g. python -m crds.newcontext now called as python -m crds.refactoring.newcontext

_REFACTORING_MODULES = [
    "checksum",
    "newcontext",
    "refactor",
    "refactor2",
]

# e.g. make crds.rmap importable same as crds.core.rmap reorganized code
_alias_subpackage_module("crds.refactoring", _REFACTORING_MODULES)

# ============================================================================

# e.g. python -m crds.uniqname now called as -m crds.refactoring.uniqname

_MISC_MODULES = [
    "datalvl",               # external interface with pipelines
    "query_affected",        # external interface with pipelines
    "uniqname",              # external interface with submitters

    "check_archive",         # misc utility
    "sql",                   # prototype convenience wrapper
]

# e.g. make crds.rmap importable same as crds.core.rmap reorganized code
_alias_subpackage_module("crds.misc", _MISC_MODULES)

# ============================================================================

URL = os.environ.get("CRDS_SERVER_URL", "https://crds-serverless-mode.stsci.edu")
api.set_crds_server(URL)



from .certify import certify_files, certify_file, CertifyScript
from crds.core import generic_tpn, reftypes  # backward compatability

__all__ = [
    "certify_files",
    "certify_file",
    "CertifyScript",
]



def tstmod(module):
    """Wrap test_config.tstmod to configure standard options throughout CRDS test_config."""
    import doctest
    # doctest.ELLIPSIS_MARKER = '-etc-'
    return doctest.testmod(module, optionflags=doctest.ELLIPSIS)



"""This module contains functions for checking for identical existing files on
the CRDS server.
"""
import os.path

from crds import client
from crds.core import utils, log
from crds.core.exceptions import DuplicateSha1sumError

def check_sha1sums(filepaths, observatory=None):
    """Given a list of filepaths which nominally will be submitted
    to CRDS for project `observatory`,  check to see if any are
    bit-for-bit identical with existing files as determined by
    the CRDS server's catalog and sha1sum matching.

    filepaths  [str, ...]   paths of files to be checked for preexistence
    observartory   str    e.g. 'hst' or 'jwst'

    Returns    count of duplicate files
    """
    log.info("Checking local file sha1sums vs. CRDS server to identify files already in CRDS.")
    sha1sums = get_all_sha1sums(observatory)
    for filepath in filepaths:
        check_sha1sum(filepath, sha1sums, observatory)

def check_sha1sum(filepath, sha1sums=None, observatory=None):
    """Check to see if the sha1sum of `filepath` is identical to any
    of the files mentioned in `sha1sums`.

    Return 1   IFF `filepath` is a duplicate of an existing CRDS file.
    Otherwise   0
    """
    if sha1sums is None:
        sha1sums = get_all_sha1sums(observatory)
    sha1sum = utils.checksum(filepath)
    log.verbose("Checking file", repr(filepath), "with sha1sum", repr(sha1sum),
                "for duplication on CRDS server.")
    if sha1sum in sha1sums:
        raise DuplicateSha1sumError(
            "File", repr(os.path.basename(filepath)),
            "is identical to existing CRDS file", repr(sha1sums[sha1sum]))

def get_all_sha1sums(observatory=None):
    """Query the CRDS server for sha1sums for all existing files.

    Returns {sha1sum : name, ...}
    """
    if observatory is None:
        observatory = client.get_default_observatory()
    fileinfo = client.get_file_info_map(
        observatory, fields=["sha1sum"])
    sha1sums = { info["sha1sum"]: name
                 for name,info in fileinfo.items() }
    log.verbose("Retrieved", len(sha1sums), "sha1sums for existing CRDS files.")
    return sha1sums
